#! /usr/bin/env python
#
"""
Manage startup and teardown of cloud-based VMs to run applications.

This module implements the cloud-independent logic as an abstract base
class, which needs to be subclassed to add the actual cloud-dependent
API calls.
"""
# Copyright (C) 2012-2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__docformat__ = 'reStructuredText'


from abc import ABCMeta, abstractmethod
import os
import os.path
import paramiko
from string import ascii_letters, digits

# GC3Pie imports
import gc3libs
import gc3libs.defaults
from gc3libs.exceptions import \
    ConfigurationError, \
    LRMSSubmitError, \
    MaximumCapacityReached, \
    ResourceNotReady, \
    TransportError, \
    UnrecoverableAuthError, \
    UnrecoverableDataStagingError, \
    UnrecoverableError
import gc3libs.url
from gc3libs import Run
from gc3libs.utils import expandpath, insert_char_every_n_chars, same_docstring_as
from gc3libs.backends import LRMS
from gc3libs.backends.shellcmd import ShellcmdLrms
from gc3libs.utils import cache_for, lookup
from gc3libs.quantity import MiB


class VmInfo(Struct):
    """
    Record data about a started VM instance.

    A `VmInfo` object is mostly a free-form record, but the
    constructor still enforces the following constraints:

    * There must be a non-empty `vmid` attribute.

    .. note::

      The `vmid` attribute must be unique among all `VmInfo` instances!

      **FIXME:** This is not currently enforced by the constructor.

    The `state` attribute is set to one of the following strings
    (defaults to `DOWN` if not given in the constructor):

    =========  ============================================================
    status     meaning
    =========  ============================================================
    STARTING   A request to start the VM has been sent to the Cloud provider,
               but the VM is not ready yet.
    CONNECTING The machine is ready to run jobs.
    READY      The machine is ready to run jobs.
    DRAINING   The VM is scheduled to stop, waiting for all batch jobs to terminate
               before issuing the 'halt' command.
    STOPPING   The 'halt' command is about to be issued to the cloud provider.
    DOWN       The VM has been stopped and cannot be restarted/resumed;
               once a VM reaches this state, it is removed from the list
               of VMs at the next `Orchestrator` cycle.
    OTHER      Unexpected/unhandled state; usually signals an error.
    =========  ============================================================


    The following attributes are available after the machine has
    reached the ``READY`` state:

    ============  ========  ================================================
    attribute     type      meaning
    ============  ========  ================================================
    ready_at      datetime  when the machine notified it's ready to run jobs
    jobs          list      list of Job IDs of jobs running on this node
    nodename      str       machine node name (as reported in the batch system listing)
    ============  ========  ================================================
    """

    STARTING = 'STARTING'
    CONNECTING = 'CONNECTING'
    READY = 'READY'
    STOPPING = 'STOPPING'
    DRAINING = 'DRAINING'
    DOWN = 'DOWN'
    OTHER = 'OTHER'

    def __init__(self, *args, **kwargs):
        Struct.__init__(self, *args, **kwargs)
        # ensure required fields are there
        assert 'vmid' in self, ("VmInfo object %s missing required field 'vmid'" % self)
        # provide defaults
        if 'state' not in self:
            self.state = VmInfo.DOWN
        if 'jobs' not in self:
            self.jobs = set()


    @property
    def state(self):
        return self._state

    @property.setter
    def state(self, value):
        assert value in [
            VmInfo.STARTING,
            VmInfo.CONNECTING,
            VmInfo.READY,
            VmInfo.DRAINING,
            VmInfo.STOPPING,
            VmInfo.DOWN,
            VmInfo.OTHER
        ]
        self._state = value


    def __str__(self):
        if 'nodename' in self:
            return ("VM Node '%s'" % self.nodename)
        else:
            return ("VM %s" % self.vmid)
    __repr__ = __str__


    def __hash__(self):
        """Use the VM id as unique hash value."""
        return hash(self.vmid)


    def is_alive(self):
        """
        Return `True` if the VM is up or will soon be (i.e., it is starting now).
        """
        return self.state in [VmInfo.STARTING, VmInfo.READY]


class CloudLrms(LRMS, ABCMeta):
    """
    Generic cloud resource.
    """
    RESOURCE_DIR = '~/.gc3/cloud.d'

    def __init__(self, name,
                 # these parameters are inherited from the `LRMS` class
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime,
                 # these are common to all cloud-based LRMSes
                 keypair_name, public_key,
                 vm_auth, vm_overhead,
                 resourcedir=None,
                 # extra args are used to instanciate "sub-resources"
                 **extra_args):

        # Note: this creates attributes from key/value pairs given in the
        # `extra_args` parameters. In particular, the `self.type` attribute
        # (referenced below) is set in this chained constructor...
        LRMS.__init__(
            self, name,
            architecture, max_cores, max_cores_per_job,
            max_memory_per_core, max_walltime, auth, **extra_args)

        self.resourcedir = expandpath(resourcedir or self.RESOURCE_DIR)

        self.free_slots = int(max_cores)
        self.user_run = 0
        self.user_queued = 0
        self.queued = 0
        self._flavors = []

        # Mapping of job.execution.instance_id => LRMS
        self.subresources = {}

        # pylint: disable=no-member
        self.subresource_type = self.type.split('+', 1)[1]
        if self.subresource_type not in available_subresource_types:
            raise UnrecoverableError("Invalid resource type: %s" % self.type)

        # Keypair names can only contain alphanumeric chars!
        if not set(keypair_name).issubset(set(ascii_letters + digits + '_-.')):
            raise ConfigurationError(
                "Keypair name `{0}` is invalid:"
                " keypair names can only consist of"
                " alphanumeric chars, plus `_`, `-`, and `.`"
                .format(keypair_name))
        self.keypair_name = keypair_name
        self.public_key = expandpath(public_key.strip())
        self.image_id = image_id
        self.user_data = user_data
        self.vm_overhead = gc3libs.quantity.Memory(vm_overhead)
        self._parse_security_group()

        self.instance_type = instance_type
        # `*_instance_type` config items should be consumed here,
        # not in any sub-resource
        for key, value in list(extra_args.items()):
            if key.endswith('_instance_type'):
                self[key] = value
                extra_args.pop(key)

        # `self.subresource_args` is used to create subresources
        self.subresource_args = extra_args
        self.subresource_args['type'] = self.subresource_type
        self.subresource_args['architecture'] = self['architecture']
        self.subresource_args['max_cores'] = self['max_cores']
        self.subresource_args['max_cores_per_job'] = self['max_cores_per_job']
        self.subresource_args['max_memory_per_core'] = \
            self['max_memory_per_core']
        self.subresource_args['max_walltime'] = self['max_walltime']
        # SSH-specific configuration
        self.subresource_args['transport'] = 'ssh'
        self.subresource_args['auth'] = vm_auth
        self.subresource_args['ssh_timeout'] = 7  # FIXME: hard-coded!
        self.subresource_args['ignore_ssh_host_keys'] = True
        self.subresource_args['keyfile'] = self.public_key
        if self.subresource_args['keyfile'].endswith('.pub'):
            self.subresource_args['keyfile'] = \
              self.subresource_args['keyfile'][:-len('.pub')]
        # ShellcmdLrms by default trusts the configuration, instead of
        # checking the real amount of memory and number of cpus, but
        # we need the real values instead.
        if self.subresource_type == gc3libs.defaults.SHELLCMD_LRMS:
            self.subresource_args['override'] = 'True'

        if image_id is None:
            raise ConfigurationError(
                "No `image_id` specified in the configuration file.")


    @abstractmethod
    def _connect(self):
        """
        Return a valid connection to the cloud provider.
        A cloud-specific exception will be raised if no valid
        connection can be initialized.

        This method *must* be idempotent: if a connection is already
        available, it should be returned without re-initializing it or
        generating unneeded network traffic.

        This is an abstract method, which should be overridden in
        derived classes to implement actual functionality.
        """
        pass


    #
    # SSH keys management
    #
    # These methods deal with SSH key files; they are pretty generic
    # in scope and should eventually be moved into `gc3lib.utils` to
    # be re-used from other modules.
    #

    @staticmethod
    def _ssh_key_std_fingerprint(key):
        """
        Compute SSH public key fingerprint like `ssh-keygen -l -f`.

        Return a string representation of the key fingerprint in
        colon-separated hex format (just like OpenSSH commands print
        it to the terminal).
        """
        return ':'.join((part.encode('hex')
                              for part in key.get_fingerprint()))


    @staticmethod
    def _ssh_key_aws_fingerprint(keypath,
                                 reader=Crypto.PublicKey.RSA.importKey):
        """
        Compute SSH public key fingerprint like AWS EC2 does.

        Return a string representation of the key fingerprint in
        colon-separated hex format (just like OpenSSH commands print
        it to the terminal).

        Argument `privkeypath` is the filesystem path to a file
        containing the private key data.
        """
        with open(privkeypath, 'r') as keydata:
            pubkey = reader(keydata).publickey()
            return insert_char_every_n_chars(
                hashlib.md5(pubkey.exportKey('DER')).hexdigest(), ':', 2)


    #
    # Keypair management
    #

    def _check_keypair(self, keypair):
        """
        Check if the given SSH key is available locally.

        Try to locate the given SSH key first among the keys loaded
        into a running ``ssh-agent`` (if any), then in the key file
        given in the ``public_key`` configuration key.
        """
        # try with SSH agent first
        gc3libs.log.debug("Checking if keypair is registered in SSH agent...")
        agent = paramiko.Agent()
        fingerprints_from_agent = [
            self._ssh_key_std_fingerprint(privkey)
            for privkey in agent.get_keys()
        ]
        if ec2_key.fingerprint in fingerprints_from_agent:
            gc3libs.log.debug("Found remote key fingerprint in SSH agent.")
            return True

        # else, try to load from file
        keyfile = self.public_key
        if keyfile.endswith('.pub'):
            keyfile = keyfile[:-4]
        else:
            gc3libs.log.warning(
                "Option `public_key` in configuration file should contain"
                " the path to a public key file (with `.pub` ending),"
                " but `%s` was found instead. Continuing anyway.",
                self.public_key)

        privkey = None
        local_fingerprints = []
        for fmt, privkey_reader, pubkey_reader in [
                ('DSS', paramiko.DSSKey.from_private_key_file,
                 Crypto.PublicKey.DSA.importKey),
                ('RSA', paramiko.RSAKey.from_private_key_file,
                 Crypto.PublicKey.RSA.importKey),
        ]:
            try:
                gc3libs.log.debug(
                    "Trying to load key file `%s` as SSH %s key...",
                    keyfile, fmt)
                privkey = privkey_reader(keyfile)
                gc3libs.log.info(
                    "Successfully loaded key file `%s` as SSH %s key.",
                    keyfile, fmt)
                # compute public key fingerprints, for comparing them
                # with the remote one
                localkey_fingerprints = [
                    # Usual SSH key fingerprint, computed like
                    # `ssh-keygen -l -f` This is used, e.g., by
                    # OpenStack's EC2 compatibility layer.
                    self._ssh_key_std_fingerprint(privkey),
                    # Amazon EC2 computes the key fingerprint in a
                    # different way, see http://blog.jbrowne.com/?p=23 and
                    # https://gist.github.com/jtriley/7270594 for details.
                    self._ssh_key_aws_fingerprint(keyfile, pubkey_reader),
                ]
            # PasswordRequiredException < SSHException so we must
            # check this first
            except paramiko.PasswordRequiredException:
                gc3libs.log.warning(
                    "Key %s is encripted with a password, so we cannot check "
                    " if it matches the remote keypair (maybe you should"
                    " start `ssh-agent`?). Continuing without consistency"
                    " check with remote fingerprint.", keyfile)
                return False
            except paramiko.SSHException as ex:
                gc3libs.log.debug(
                    "File `%s` is not a valid %s private key: %s",
                    keyfile, fmt, ex)
                # try with next fmt
                continue
        if privkey is None:
            raise ValueError("Public key `{0}` is not a valid SSH key"
                             .format(self.public_key))

        # check key fingerprint
        if (local_fingerprints
                and ec2_key.fingerprint not in localkey_fingerprints):
            raise UnrecoverableAuthError(
                "Keypair `%s` exists but has different fingerprint than"
                " local one: local public key file `%s` has fingerprint(s)"
                " `%s`, whereas cloud API reports fingerprint `%s`."
                " Aborting!" % (
                    self.public_key,
                    self.keypair_name,
                    '/'.join(localkey_fingerprints),
                    ec2_key.fingerprint,
                ),
                do_log=True)


    @abstractmethod
    def _create_keypair(self, keypair_name, public_key_data):
        """
        Upload public key to the cloud provider under the given name.

        The public key is given as a byte string (typically the
        contents of some file).
        """
        pass


    # FIXME: we should not cache negative results (this class uses
    # `self._getkeypair()` to test for a keypair's existentce, but
    # positive ones can be cached for a longer time -- it is unlikely
    # that keypairs are deleted while a GC3Pie script is running.
    @abstractmethod
    @cache_for(120)
    def _get_keypair(self, keypair_name):
        """
        Return key pair object with the given name.
        If no key pair with that name is found, raise a ``NotFound``
        error.

        A key pair object is a cloud-specific opaque object; all
        manipulation within GC3Pie code should be done using the
        ``CloudLrms._*_keypair`` methods.
        """
        pass


    def _import_keypair(self, keypair_name, public_key_file):
        """
        Upload data in *public_key_file* as a keypair with the given name.
        """
        try:
            with open(expandpath(public_key_file)) as stream:
                key_data = fd.read()
            self._create_keypair(keypair_name, key_data)
            gc3libs.log.info(
                "Successfully uploaded key file `%s` as keypair `%s`",
                public_key_file, keypair.fingerprint, keypair_name)
            keypair = self._get_keypair(keypair_name)
            return keypair
        except Exception as err:
            raise UnrecoverableError(
                "Could not upload public key `{public_key_file}`"
                " as keypair {keypair_name}: {err}"
                .format(**locals()))


    def _create_instance(self, image_id, name='gc3pie-worker',
                         instance_type=None, user_data=None):
        """
        Create an instance using the image `image_id` and instance
        type `instance_type`. If no `instance_type` is defined, use
        the default.

        This method will also setup the keypair and the security
        groups, if needed.
        """

        args = {}
        if user_data:
            args['userdata'] = user_data

        # Check if the desired keypair is present
        try:
            keypair = self._get_keypair(self.keypair_name)
        except NotFound:
            gc3libs.log.info(
                "Keypair `%s` not found: creating it from public key file `%s`",
                self.keypair_name, self.public_key)
            # create keypair if it does not exist
            self._import_keypair(self.keypair_name, self.public_key)
        else:
            self._have_keypair(keypair)
        instance_type = instance_type or self.instance_type
        # Setup security groups
        if 'security_group_name' in self:
            self._setup_security_groups()
            args['security_groups'] = [self.security_group_name]

        # FIXME: we should add check/creation of proper security
        # groups

        nics = None
        if hasattr(self,'network_ids') and self.network_ids:
            nics=[{'net-id': netid.strip(), 'v4-fixed-ip': ''}
                  for netid in self.network_ids.split(',')]
            gc3libs.log.debug("Specifying networks for vm %s: %s",
                      name, ', '.join([nic['net-id'] for nic in nics]))
        args['nics'] = nics

        gc3libs.log.debug("Create new VM using image id `%s`", image_id)
        try:
            vm = self.compute_client.servers.create(name, image_id, instance_type,
                                            key_name=self.keypair_name, **args)
        except Exception as err:
            # scrape actual error kind and message out of the
            # exception; we do this mostly for sensible logging, but
            # could be an actual improvement to Boto to provide
            # different exception classes based on the <Code>
            # element...
            # XXX: is there a more robust way of doing this?
            # fall back to normal reporting...
            raise UnrecoverableError("Error starting instance: %s" % err)

        self._vmpool.add_vm(vm)
        gc3libs.log.info(
            "VM with id `%s` has been created and is in %s state.",
            vm.id, vm.status)
        return vm

    def _get_subresource(self, vm):
        """
        Return the resource associated to the virtual machine `vm`.

        Updates the internal list of available resources if needed.
        """
        self._connect()
        if vm.id not in self.subresources:
            self.subresources[vm.id] = self._make_subresource(
                vm.id, vm.preferred_ip)
            # Update resource based on flavor specs
            try:
                flavor = self.compute_client.flavors.get(vm.flavor['id'])
                res = self.subresources[vm.id]
                res['max_memory_per_core'] = flavor.ram * MiB
                res['max_cores_per_job'] = flavor.vcpus
                res['max_cores'] = flavor.vcpus
            except Exception as ex:
                # Ignore any error here, as we can get information on
                # the subresources when we connect to it.
                gc3libs.log.info(
                    "Ignoring error in setting max_cores_per_job/max_memory"
                    " values for new subresource %s based on flavor: %s",
                    self.subresources[vm.id].name, ex)
        return self.subresources[vm.id]

    def _get_vm(self, vm_id):
        """
        Return the instance with id `vm_id`, raises an error if there
        is no such instance with that id.
        """
        self._connect()
        vm = self._vmpool.get_vm(vm_id)
        return vm

    @cache_for(120)
    def _get_security_groups(self):
        self._connect()
        try:
            # python-novaclient < 8.0.0
            return self.compute_client.security_groups.list()
        except AttributeError:
            return self.network_client.list_security_groups()['security_groups']


    @cache_for(120)
    def _get_security_group(self, name):
        groups = self._get_security_groups()
        group = [grp for grp in groups if lookup(grp, 'name') == name]
        if not group:
            raise NotFound("Security group %s not found." % name)
        return group[0]

    def _make_subresource(self, id, remote_ip):
        """
        Create a resource associated to the instance with `remote_ip`
        ip using configuration file parameters.
        """
        gc3libs.log.debug(
            "Creating remote ShellcmdLrms resource for IP address %s",
            remote_ip)
        args = self.subresource_args.copy()
        args['frontend'] = remote_ip
        args['name'] = "%s@%s" % (id, self.name)
        return ShellcmdLrms(**args)

    def _parse_security_group(self):
        """
        Parse configuration file and set `self.security_group_rules`
        with a list of dictionaries containing the rule sets
        """
        rules = self.security_group_rules.split(',')
        self.security_group_rules = []
        for rule in rules:
            rulesplit = rule.strip().split(':')
            if len(rulesplit) != 4:
                gc3libs.log.warning("Invalid rule specification in"
                                    " `security_group_rules`: %s" % rule)
                continue
            self.security_group_rules.append({
                'ip_protocol': rulesplit[0],
                'from_port': int(rulesplit[1]),
                'to_port': int(rulesplit[2]),
                'cidr_ip': rulesplit[3],
            })

    def _setup_security_groups(self):
        """
        Check the current configuration and set up the security group
        if it does not exist.
        """
        self._connect()
        if not self.security_group_name:
            gc3libs.log.error("Group name in `security_group_name`"
                              " configuration option cannot be empty!")
            return

        try:
            self._get_security_group(self.security_group_name)
        except NotFound:
            try:
                gc3libs.log.info("Creating security group %s",
                                 self.security_group_name)

                self.compute_client.security_groups.create(
                    self.security_group_name,
                    "GC3Pie_%s" % self.security_group_name)
            except Exception as ex:
                gc3libs.log.error("Error creating security group %s: %s",
                                  self.security_group_name, ex)
                raise UnrecoverableError(
                    "Error creating security group %s: %s"
                    % (self.security_group_name, ex))

            self._get_security_group(self.security_group_name)
        # TODO: Check if the security group has all the rules we want
        # security_group = groups[self.security_group_name]
        # current_rules = []
        # for rule in security_group.rules:
        #     rule_dict = {
        #         'ip_protocol':  rule.ip_protocol,
        #         'from_port':    int(rule.from_port),
        #         'to_port':      int(rule.to_port),
        #         'cidr_ip':      str(rule.grants[0]),
        #     }
        #     current_rules.append(rule_dict)

        # for new_rule in self.security_group_rules:
        #     if new_rule not in current_rules:
        #         security_group.authorize(**new_rule)

    @cache_for(120)
    def _get_available_images(self):
        self._connect()
        try:
            # python-novaclient < 8.0.0
            return self.compute_client.images.list()
        except AttributeError:
            # ``glance_client.images.list()`` returns a generator, but callers
            # of `._get_available_images()` expect a Python list
            return list(self.image_client.images.list())

    @cache_for(120)
    def _get_available_flavors(self):
        self._connect()
        return self.compute_client.flavors.list()

    def _get_flavor(self, name):
        for flavor in self._flavors:
            # pick the first match by that name
            if flavor.name == name:
                return flavor
        # no flavor by the given name
        raise NotFound("Flavor `{name}` not found.".format(name=name))

    @cache_for(120)
    def _get_keypair(self, keypair_name):
        self._connect()
        return self.compute_client.keypairs.get(keypair_name)

    def get_image_id_for_job(self, job):
        """
        If a configuration option <application>_image_id is present,
        returns its value, otherwise returns `self.image_id`
        """
        conf_option = job.application_name + '_image_id'
        if conf_option in self:
            return self[conf_option]
        else:
            return self.image_id

    def get_instance_type_for_job(self, task):
        """
        Return the "best" instance_type (aka flavor) for the task.

        This method will always returns an instance_type that accomodates the
        task's requirements. Preference will be given, in order, to:

        1) ``<application>_instance_type configuration`` option
        2) ``instance_type configuration`` option
        3) the smallest flavor (ordering by cpus, ram and disk)
           that fits the application requirements.
        """
        req_cores = self._get_task_requirement(task, 'requested_cores', 1)
        req_memory = self._get_task_requirement(task, 'requested_memory', 0*MiB)

        # If there is an option <application>_instance_type, try to use it
        for conf_option in [
                task.application_name + '_instance_type',
                'instance_type'
        ]:
            gc3libs.log.debug("Trying configuration option `%s` ...", conf_option)
            try:
                flavor_name = self[conf_option]
            except KeyError:
                # ignore and try next option
                continue
            if flavor_name is None:
                continue

            try:
                flavor = self._get_flavor(flavor_name)
            except NotFound:
                gc3libs.log.warning(
                    "Unable to use flavor `%s` for task `%s` as requested"
                    " by config option `%s` because it's not available"
                    " in the cloud backend. Fix configuration file?",
                    flavor_name, task, conf_option)
                continue

            if self._flavor_matches_requirements(flavor, req_cores, req_memory):
                gc3libs.log.info(
                    "Using flavor `%s` for task `%s`"
                    " as requested by config option `%s`",
                    flavor.name, task, conf_option)
                return flavor
            else:
                gc3libs.log.warning(
                    "Unable to use flavor `%s` for task `%s`"
                    " as requested by config option `%s`"
                    " because it doesn't fit application's requirements",
                    flavor_name, task, conf_option)
                continue

        # fall back to smallest flavor that fits
        valid_flavors = [
            flv for flv in self._flavors
            if self._flavor_matches_requirements(flv, req_cores, req_memory)
        ]
        flavor = min(valid_flavors, key=self._flavor_ordering_key)
        gc3libs.log.info(
            "Using VM flavor `%s` as the smallest flavor that fits requirements"
            " of task `%s`", flavor.name, task)
        return flavor

    @staticmethod
    def _get_task_requirement(task, reqname, default):
        """
        Return attribute `reqname` of `task`.

        If the attribute is not set or is ``None``, then return `default`
        instead.
        """
        return getattr(task, reqname, default)

    def _flavor_matches_requirements(self, flavor, req_cores, req_memory):
        """
        Return whether the flavor matches the given requirements.
        """
        return (flavor.vcpus >= req_cores and
                (flavor.ram * MiB - self.vm_os_overhead) >= req_memory)

    @staticmethod
    def _flavor_ordering_key(flavor):
        """
        Return the value to use when sorting flavors.

        By default, flavors are sorted lexicographically by:
        - nr. of vCPUS (first),
        - amount of RAM,
        - size of disk (last).
        """
        return (flavor.vcpus, flavor.ram, flavor.disk)


    def get_user_data_for_job(self, job):
        """
        If a configuration option <application>_user_data is present,
        returns its value, otherwise returns `self.user_data`.
        """
        conf_option = job.application_name + '_user_data'
        if conf_option in self:
            return self[conf_option]
        else:
            return self.user_data

    @same_docstring_as(LRMS.cancel_job)
    def cancel_job(self, app):
        try:
            subresource = self._get_subresource(
                self._get_vm(app.execution._lrms_vm_id))
            return subresource.cancel_job(app)
        except InstanceNotFound:
            # ignore -- if this VM exists no more, we need not cancel any job
            pass
        except UnrecoverableError as err:
            gc3libs.log.error(
                "Changing state of task '%s' to UNKNOWN because of"
                " an OpenStack API error (%s: %s)",
                app.execution.lrms_jobid, err.__class__.__name__, err)
            app.execution.state = Run.State.UNKNOWN
            app.execution.history.append(
                "State changed to UNKNOWN because of"
                " an OpenStack API error (%s: %s)."
                % (err.__class__.__name__, err))
            raise

    @same_docstring_as(LRMS.get_resource_status)
    def get_resource_status(self):
        self.updated = False
        # Since we create the resource *before* the VM is actually up
        # & running, it's possible that the `frontend` value of the
        # resources points to a non-existent hostname. Therefore, we
        # have to update them with valid public_ip, if they are
        # present.

        # Update status of known VMs
        for vm_id in self._vmpool:
            try:
                vm = self._vmpool.get_vm(vm_id, force_reload=True)
            except UnrecoverableError as ex:
                gc3libs.log.warning(
                    "Removing stale information on VM `%s`. It has probably"
                    " been deleted from outside GC3Pie.", vm_id)
                self._vmpool.remove_vm(vm_id)
                continue

            if vm.status in PENDING_STATES:
                # If VM is still in pending state, skip creation of
                # the resource
                continue
            elif vm.status in ERROR_STATES:
                # The VM is in error state: exit.
                gc3libs.log.error(
                    "VM with id `%s` is in ERROR state."
                    " Terminating it!", vm.id)
                vm.delete()
                self._vmpool.remove_vm(vm.id)
                self.subresources.pop(vm.id)
                continue
            elif vm.status == 'DELETED':
                gc3libs.log.info(
                    "VM `%s` in DELETED state. It has probably been terminated"
                    " from outside GC3Pie. Removing it from the list of VMs.",
                    vm.id)
                self._vmpool.remove_vm(vm.id)
                self.subresources.pop(vm.id)
                continue
            elif vm.status in ['SHUTOFF', 'SUSPENDED',
                               'RESCUE', 'VERIFY_RESIZE']:
                # The VM has probably ben stopped or shut down from
                # outside GC3Pie.
                gc3libs.log.error(
                    "VM with id `%s` is in permanent state `%s`.",
                    vm.id, vm.status)
                continue

            # Get or create a resource associated to the vm
            subresource = self._get_subresource(vm)
            try:
                subresource.get_resource_status()
            except TransportError as ex:
                # TODO: get all the IPs and try with all of them to connect.
                # Start with preferred_ip if defined
                gc3libs.log.info(
                    "Ignoring error in updating resource '%s': %s."
                    " Trying other IPs.", subresource.name, ex)
                for ip in sum(list(vm.networks.values()), []):
                    if vm.preferred_ip == ip:
                        continue
                    vm.preferred_ip = ip
                    subresource.frontend = ip
                    gc3libs.log.info(
                        "Connection error. Trying with alternate IP address "
                        "%s", vm.preferred_ip)
                    try:
                        subresource.get_resource_status()
                        break
                    except Exception as ex:
                        gc3libs.log.info(
                            "Ignoring error in updating resource '%s': %s."
                            " The corresponding VM may not be ready yet.",
                            subresource.name, ex)
                # Unable to connect to the VM using any IP.  Ensure
                # this resource is considered "pending" as we couldn't
                # update its status
                subresource.updated = False
            except Exception as ex:
                # XXX: Actually, we should try to identify the kind of
                # error we are getting. For instance, if the
                # configuration options `username` is wrong, we will
                # create VMs but we will never be able to submit jobs
                # to them, thus causing an increasing number of
                # useless VMs created on the cloud.
                if gc3libs.error_ignored(
                        # context:
                        # - module
                        'openstack',
                        # - class
                        'OpenStackLrms',
                        # - method
                        'get_resource_status',
                        # - actual error class
                        ex.__class__.__name__,
                        # - additional keywords
                        'resource',
                        'status',
                        'update',
                        'vm',
                ):
                    gc3libs.log.info(
                        "Ignoring error while updating resource '%s'. "
                        "The corresponding VM may not be ready yet. Error: %s",
                        subresource.name, ex)
                else:
                    # propagate exception back to caller
                    raise
        self._vmpool.update()
        return self

    @same_docstring_as(LRMS.get_results)
    def get_results(self, app, download_dir, overwrite=False,
                    changed_only=True):
        try:
            subresource = self._get_subresource(
                self._get_vm(app.execution._lrms_vm_id))
        except InstanceNotFound:
            gc3libs.log.error(
                "Changing state of task '%s' to TERMINATED since OpenStack"
                " instance '%s' does not exist anymore.",
                app.execution.lrms_jobid, app.execution._lrms_vm_id)
            app.execution.state = Run.State.TERMINATED
            app.execution.signal = Run.Signals.RemoteError
            app.execution.history.append(
                "State changed to TERMINATED since OpenStack"
                " instance '%s' does not exist anymore."
                % (app.execution._lrms_vm_id,))
            raise UnrecoverableDataStagingError(
                "VM where job was running is no longer available")
        except UnrecoverableError as err:
            gc3libs.log.error(
                "Changing state of task '%s' to UNKNOWN because of"
                " an OpenStack API error (%s: %s)",
                app.execution.lrms_jobid, err.__class__.__name__, err)
            app.execution.state = Run.State.UNKNOWN
            app.execution.history.append(
                "State changed to UNKNOWN because of"
                " an OpenStack API error (%s: %s)."
                % (err.__class__.__name__, err))
            raise
        return subresource.get_results(
            app, download_dir, overwrite, changed_only)

    @same_docstring_as(LRMS.update_job_state)
    def update_job_state(self, app):
        self._connect()
        if app.execution._lrms_vm_id not in self.subresources:
            try:
                self.subresources[app.execution._lrms_vm_id] = self._get_subresource(
                    self._get_vm(app.execution._lrms_vm_id))
            except InstanceNotFound:
                gc3libs.log.error(
                    "Changing state of task '%s' to TERMINATED since OpenStack"
                    " instance '%s' does not exist anymore.",
                    app.execution.lrms_jobid, app.execution._lrms_vm_id)
                app.execution.state = Run.State.TERMINATED
                app.execution.signal = Run.Signals.RemoteError
                app.execution.history.append(
                    "State changed to TERMINATED since OpenStack"
                    " instance '%s' does not exist anymore."
                    % (app.execution._lrms_vm_id,))
                raise
            except UnrecoverableError as err:
                gc3libs.log.error(
                    "Changing state of task '%s' to UNKNOWN because of"
                    " an OpenStack API error (%s: %s)",
                    app.execution.lrms_jobid, err.__class__.__name__, err)
                app.execution.state = Run.State.UNKNOWN
                app.execution.history.append(
                    "State changed to UNKNOWN because of"
                    " an OpenStack API error (%s: %s)."
                    % (err.__class__.__name__, err))
                raise
        return self.subresources[app.execution._lrms_vm_id].update_job_state(app)

    def submit_job(self, job):
        """
        Submission on an OpenStack resource will usually happen in
        multiple steps, since creating a VM and attaching a resource
        to it will take some time.

        In order to return as soon as possible, the backend will raise
        a `RecoverableError` whenever submission is delayed.

        In case a permanent error is found (for instance, we cannot
        create VMs on the cloud), a `UnrecoverableError` is raised.

        More in detail, when during submission the following will
        happen:

        * First of all, the backend will try to submit the job to one
          of the already available subresources.

        * If none of them is able to sbmit the job, the backend will
          check if there is a VM in pending state, and in case there
          is one it will raise a `RecoverableError`, thus delaying
          submission.

        * If no VM in pending state is found, the `vm_pool_max_size`
          configuration option is checked. If we already reached the
          maximum number of VM, a `UnrecoverableError` is raised.

        * If no VM in pending state is found but `vm_pool_max_size` is
          still lesser than the number of VM currently created (or it
          is None, which for us means no limit), then a new VM is
          created, and `RecoverableError` is raised.

        """
        # Updating resource is needed to update the subresources. This
        # is not always done before the submit_job because of issue
        # nr.  386:
        #     https://github.com/uzh/gc3pie/issues/386
        self.get_resource_status()
        pending_vms = set(vm.id for vm in self._vmpool.get_all_vms()
                          if vm.status in PENDING_STATES)

        image_id = self.get_image_id_for_job(job)
        # Check if the image id is valid
        if image_id not in [img.id for img in self._get_available_images()]:
            raise ConfigurationError("Image ID %s not found in cloud "
                                     "%s" % (image_id, self.os_auth_url))

        instance_type = self.get_instance_type_for_job(job)
        if not instance_type:
            raise RuntimeError(
                "Unable to find a suitable instance type for "
                "application %s" % job)

        # First of all, try to submit to one of the subresources.
        for vm_id, subresource in list(self.subresources.items()):
            if not subresource.updated:
                # The VM is probably still booting, let's skip to the
                # next one and add it to the list of "pending" VMs.
                pending_vms.add(vm_id)
                continue
            try:
                # Check that the required image id and instance type
                # are correct
                vm = self._get_vm(vm_id)
                if vm.image['id'] != image_id:
                    continue
                subresource.submit_job(job)
                job.execution._lrms_vm_id = vm_id
                job.changed = True
                gc3libs.log.info(
                    "Job successfully submitted to remote resource %s.",
                    subresource.name)
                return job
            except (LRMSSubmitError, InstanceNotFound) as ex:
                if gc3libs.error_ignored(
                        # context:
                        # - module
                        'openstack',
                        # - class
                        'OpenStackLrms',
                        # - method
                        'submit_job',
                        # - actual error class
                        ex.__class__.__name__,
                        # - additional keywords
                        'submit',
                ):
                    gc3libs.log.debug(
                        "Ignoring error in submitting to resource '%s': %s",
                        subresource.name, ex)
                else:
                    # propagate error back to caller
                    raise

        # Couldn't submit to any resource.
        if not pending_vms:
            # No pending VM, and no resource available. Create a new VM
            if not self.vm_pool_max_size \
                    or len(self._vmpool) < self.vm_pool_max_size:
                user_data = self.get_user_data_for_job(job)
                vm = self._create_instance(
                    image_id,
                    name="GC3Pie_%s_%d" % (self.name, (len(self._vmpool) + 1)),
                    instance_type=instance_type,
                    user_data=user_data)
                pending_vms.add(vm.id)

                self._vmpool.add_vm(vm)
            else:
                raise MaximumCapacityReached(
                    "Already running the maximum number of VM on resource %s:"
                    " %d VMs started, but max %d allowed by configuration."
                    % (self.name, len(self._vmpool), self.vm_pool_max_size),
                    do_log=True)

        # If we reached this point, we are waiting for a VM to be
        # ready, so delay the submission until we wither can submit to
        # one of the available resources or until all the VMs are
        # ready.
        gc3libs.log.debug(
            "No available resource was found, but some VM is still in"
            " `pending` state. Waiting until the next iteration before"
            " creating a new VM. Pending VM ids: %s", pending_vms)
        raise ResourceNotReady(
            "Delaying submission until one of the VMs currently pending"
            " is ready. (Pending VM ids: %r)" % (pending_vms,))

    @same_docstring_as(LRMS.peek)
    def peek(self, app, remote_filename, local_file, offset=0, size=None):
        try:
            subresource = self._get_subresource(
                self._get_vm(app.execution._lrms_vm_id))
        except InstanceNotFound:
            gc3libs.log.error(
                "Changing state of task '%s' to TERMINATED since OpenStack"
                " instance '%s' does not exist anymore.",
                app.execution.lrms_jobid, app.execution._lrms_vm_id)
            app.execution.state = Run.State.TERMINATED
            app.execution.signal = Run.Signals.RemoteError
            app.execution.history.append(
                "State changed to TERMINATED since OpenStack"
                " instance '%s' does not exist anymore."
                % (app.execution._lrms_vm_id,))
            raise UnrecoverableDataStagingError(
                "VM where job was running is no longer available.")
        return subresource.peek(app, remote_filename, local_file, offset, size)

    def validate_data(self, data_file_list=None):
        """
        Supported protocols: file
        """
        for url in data_file_list:
            if url.scheme not in ['file', 'http', 'https', 'swift', 'swifts', 'swt', 'swts']:
                return False
        return True

    def free(self, app):
        """
        Free up any remote resources used for the execution of `app`.
        In particular, this should terminate any remote instance used
        by the job.

        Call this method when `app.execution.state` is anything other
        than `TERMINATED` results in undefined behavior and will
        likely be the cause of errors later on.  Be cautious.
        """
        # XXX: this should probably done only when no other VMs are
        # using this resource.

        # FIXME: freeing the resource from the application is probably
        # not needed since instances are not persistent.

        # freeing the resource from the application is now needed as
        # the same instanc may run multiple applications
        try:
            subresource = self._get_subresource(
                self._get_vm(app.execution._lrms_vm_id))
        except InstanceNotFound:
            # ignore -- if the instance is no more, there is
            # nothing we should free
            return
        subresource.free(app)

        # FIXME: current approach in terminating running instances:
        # if no more applications are currently running, turn the instance off
        # check with the associated resource
        subresource.get_resource_status()
        if not subresource.has_running_tasks():
            # turn VM off
            vm = self._get_vm(app.execution._lrms_vm_id)

            gc3libs.log.info("VM instance %s at %s is no longer needed."
                             " Terminating.", vm.id, vm.preferred_ip)
            del self.subresources[vm.id]
            vm.delete()
            del self._vmpool[vm.id]

    @same_docstring_as(LRMS.close)
    def close(self):
        gc3libs.log.info("Closing connection to cloud '%s'...",
                         self.name)
        if not self.enabled:
            # The resources was most probably disabled by command
            # line. We didn't update it before, so we don't care about
            # currently running VMs now.
            return
        # Update status of VMs and remote resources
        self.get_resource_status()
        for vm_id, subresource in list(self.subresources.items()):
            if subresource.updated and not subresource.has_running_tasks():
                vm = self._get_vm(vm_id)
                gc3libs.log.warning(
                    "VM instance %s at %s is no longer needed. "
                    "You may need to terminate it manually.",
                    vm.id, vm.preferred_ip)
                self._vmpool.remove_vm(vm.id)
            subresource.close()

    def __getstate__(self):
        # Do not save `novaclient.client.Client` class as it is
        # not pickle-compliant
        state = self.__dict__.copy()
        del state['client']
        del state['__connected']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.compute_client, self.image_client, self.network_client = self._new_clients()
        self._connect()



#!/usr/bin/env python

"""
Top-level classes for task execution and control.
"""
# Copyright (C) 2009-2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 2110-1301 USA


from __future__ import absolute_import, print_function, unicode_literals
from builtins import next
from builtins import filter
from builtins import str
from builtins import range
from builtins import object
from collections import defaultdict, deque
from fnmatch import fnmatch
import functools
import itertools
import os
import posix
import sys
import time
import tempfile
from warnings import warn

from dictproxyhack import dictproxy

import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.events import TaskStateChange, TermStatusChange
import gc3libs.exceptions
from gc3libs.quantity import Duration
import gc3libs.utils as utils


__docformat__ = 'reStructuredText'


class MatchMaker(object):
    """
    Select and sort resources for attempting submission of a `Task`.

    A match-making algorithm must implement two methods:

    - `filter`: given a task and a list of resources, return the list
      of resources that the given task could be submitted to.

    - `rank`: given a task and a list of resources, return a list of
      resources sorted in preference order, i.e., submission of the
      given task will be attempted to the first returned resource,
      then the next one, etc.

    This class implements the default match-making algorithm in
    GC3Pie, which operates as follows:

    - *filter phase:* if `task` has a `compatible_resources` method (as
      instances of `Application`:class: do), retain only those
      resources where it evaluates to ``True``.  Otherwise, return the
      resources list unchanged.

    - *rank phase:* sort resources according to the task's
      `rank_resources` method, or retain the given order if task does
      not define such method.
    """

    # pylint: disable=no-self-use
    def filter(self, task, resources):
        """
        Return the subset of resources to which `task` could be submitted to.

        Note that the result subset could be empty (no resource can
        accomodate task's requirements).

        The default implementation uses the task's
        `compatible_resources` method to retain only the resources
        that satisfy the task's requirements.  If `task` does not
        provide such a method, the resource list is returned
        unchanged.
        """
        gc3libs.log.debug(
            "Performing matching of resource(s) %s to task '%s' ...",
            ','.join(r.name for r in resources), task)
        # keep only compatible resources
        try:
            compatible_resources = task.compatible_resources(resources)
            gc3libs.log.debug(
                'Task compatiblity check returned %d matching resources',
                len(compatible_resources))
        except AttributeError:
            # XXX: should we require that len(resources) > 0?
            compatible_resources = resources
        return compatible_resources

    # pylint: disable=no-self-use
    def rank(self, task, resources):
        """
        Sort the list of `resources` in the preferred order for submitting
        `task`.

        Unless overridden in a derived class, this calls the task's
        `rank_resources` method to sort the list.  If the task does
        not provide such a method, the resources list is returned
        unchanged.
        """
        # sort resources according to the Task's preference, if stated
        try:
            targets = task.rank_resources(resources)
        except AttributeError:
            targets = resources
        return targets


class Core(object):
    """
    Core operations: submit, update state, retrieve (a snapshot of) output,
    cancel job.

    Core operations are *blocking*, i.e., they return only after the
    operation has successfully completed, or an error has been detected.

    Operations are always performed by a `Core` object.  `Core` implements
    an overlay Grid on the resources specified in the configuration file.

    Initialization of a `Core`:class: instance also initializes all
    resources in the passed `Configuration`:class: instance.  By default,
    GC3Pie's `Core` objects will ignore errors in initializing resources,
    and only raise an exception if *no* resources can be initialized.
    This can be changed by either passing an optional argument
    ``resource_errors_are_fatal=True``, or by setting the environmental
    variable ``GC3PIE_RESOURCE_INIT_ERRORS_ARE_FATAL`` to ``yes`` or ``1``.
    """

    def __init__(self, cfg, matchmaker=MatchMaker(),
                 resource_errors_are_fatal=None):
        # propagate resource init errors?
        if resource_errors_are_fatal is None:
            # get result from the environment
            resource_errors_are_fatal = gc3libs.utils.string_to_boolean(
                os.environ.get('GC3PIE_RESOURCE_INIT_ERRORS_ARE_FATAL', 'no'))

        # init auths
        self.auto_enable_auth = cfg.auto_enable_auth

        # init backends
        self.resources = cfg.make_resources(
            ignore_errors=(not resource_errors_are_fatal))
        if len(self.resources) == 0:
            raise gc3libs.exceptions.NoResources(
                "No resources given to initialize `gc3libs.core.Core` object!")

        # init matchmaker
        self.matchmaker = matchmaker

    def get_backend(self, name):
        try:
            return self.resources[name]
        except KeyError:
            raise gc3libs.exceptions.InvalidResourceName(
                "No configured resource by the name '%s'"
                % (name,))

    def select_resource(self, match):
        """
        Disable resources that *do not* satisfy predicate `match`.
        Return number of enabled resources.

        Argument `match` can be:

          - either a function (or a generic callable) that is passed
            each `Resource` object in turn, and should return a
            boolean indicating whether the resources should be kept
            (`True`) or not (`False`);

          - or it can be a string: only resources whose name matches
            (wildcards ``*`` and ``?`` are allowed) are retained.

        .. note::

          Calling this method modifies the configured list of
          resources in-place.
        """
        enabled = 0
        for lrms in self.resources.values():
            try:
                if not match(lrms):
                    lrms.enabled = False
            # we expect `TypeError: 'str' object is not callable` in case
            # argument `match` is a string
            except TypeError:
                if not fnmatch(lrms.name, match):
                    lrms.enabled = False
            if lrms.enabled:
                enabled += 1
        return enabled

    def free(self, app, **extra_args):
        """
        Free up any remote resources used for the execution of `app`.
        In particular, this should delete any remote directories and
        files.

        It is an error to call this method if `app.execution.state` is
        anything other than `TERMINATED`: an `InvalidOperation` exception
        will be raised in this case.

        :raise: `gc3libs.exceptions.InvalidOperation` if `app.execution.state`
                differs from `Run.State.TERMINATED`.
        """
        assert isinstance(
            app, Task), "Core.free: passed an `app` argument which" \
            " is not a `Task` instance."
        if isinstance(app, Application):
            return self.__free_application(app, **extra_args)
        else:
            # must be a `Task` instance
            return self.__free_task(app, **extra_args)

    # pylint: disable=unused-argument
    def __free_application(self, app, **extra_args):
        """Implementation of `free` on `Application` objects."""
        if app.execution.state not in [
                Run.State.TERMINATING, Run.State.TERMINATED]:
            raise gc3libs.exceptions.InvalidOperation(
                "Attempting to free resources of job '%s',"
                " which is in non-terminal state." % app)

        # auto_enable_auth = extra_args.get(
        #     'auto_enable_auth', self.auto_enable_auth)

        try:
            lrms = self.get_backend(app.execution.resource_name)
            lrms.free(app)
        except AttributeError:
            gc3libs.log.debug(
                "Core.__free_application():"
                " Application `%s` is missing the `execution.resource_name` attribute."
                " This should not happen. I'm assuming the application had been"
                " aborted before submission.",
                app)

    # pylint: disable=unused-argument
    def __free_task(self, task, **extra_args):
        """Implementation of `free` on generic `Task` objects."""
        return task.free(**extra_args)

    def submit(self, app, resubmit=False, targets=None, **extra_args):
        """Submit a job running an instance of the given task `app`.

        Upon successful submission, call the `submitted` method on the
        `app` object.  If `targets` are given, submission of the task
        is attempted to the resources in the order given; the `submit`
        method returns after the first successful attempt.  If
        `targets` is ``None`` (default), a brokering procedure is run
        to determine the best resource among the configured ones.

        At the beginning of the submission process, the
        `app.execution` state is reset to ``NEW``; if submission is
        successful, the task will be in ``SUBMITTED`` or ``RUNNING``
        state when this call returns.

        :raise: `gc3libs.exceptions.InputFileError` if an input file
                does not exist or cannot otherwise be read.

        :param Task app:
          A GC3Pie `Task`:class: instance to be submitted.
        :param resubmit:
          If ``True``, submit task regardless of its execution state;
          if ``False`` (default), submission is a no-op if task is not
          in ``NEW`` state.
        :param targets:
          A list of `Resource`s to submit the task to; resources are
          tried in the order given.  If ``None`` (default), perform
          brokering among all the configured resources.

        """
        assert isinstance(
            app, Task), "Core.submit: passed an `app` argument" \
            "which is not a `Task` instance."
        if isinstance(app, Application):
            return self.__submit_application(
                app, resubmit, targets, **extra_args)
        else:
            # must be a `Task` instance
            return self.__submit_task(app, resubmit, targets, **extra_args)

    def __submit_application(self, app, resubmit, targets, **extra_args):
        """Implementation of `submit` on `Application` objects."""

        gc3libs.log.debug("Submitting %s ...", app)

        # auto_enable_auth = extra_args.get(
        #     'auto_enable_auth', self.auto_enable_auth)

        job = app.execution
        if resubmit:
            job.state = Run.State.NEW
        elif job.state != Run.State.NEW:
            return

        # Validate Application local input files
        for input_ref in app.inputs:
            if input_ref.scheme == 'file':
                # Local file, check existence before proceeding
                if not os.path.exists(input_ref.path):
                    raise gc3libs.exceptions.UnrecoverableDataStagingError(
                        "Input file '%s' does not exist" % input_ref.path,
                        do_log=True)

        if targets is not None:
            assert len(targets) > 0
        else:  # targets is None
            enabled_resources = [
                r for r in self.resources.values() if r.enabled]
            if len(enabled_resources) == 0:
                raise gc3libs.exceptions.NoResources(
                    "Could not initialize any computational resource"
                    " - please check log and configuration file.")

            # decide which resource to use
            compatible_resources = self.matchmaker.filter(
                app, enabled_resources)
            if len(compatible_resources) == 0:
                raise gc3libs.exceptions.NoResources(
                    "No available resource can accomodate the application"
                    " requirements")
            gc3libs.log.debug(
                "Application compatibility check returned %d matching"
                " resources", len(compatible_resources))

            if len(compatible_resources) <= 1:
                # shortcut: no brokering to do, just use what we've got
                targets = compatible_resources
            else:
                # update status of selected resources
                self.update_resources(compatible_resources)
                updated_resources = [r for r in compatible_resources if r.updated]
                if len(updated_resources) == 0:
                    raise gc3libs.exceptions.LRMSSubmitError(
                        "No computational resource found reachable during"
                        " update! Aborting submission of task '%s'" %
                        app)

                # sort resources according to Application's preferences
                targets = self.matchmaker.rank(app, updated_resources)

        exs = []
        # after brokering we have a sorted list of valid resource
        for resource in targets:
            gc3libs.log.debug("Attempting submission to resource '%s'...",
                              resource.name)
            try:
                job.timestamp[Run.State.NEW] = time.time()
                job.info = ("Submitting to '%s'" % (resource.name,))
                resource.submit_job(app)
            except gc3libs.exceptions.LRMSSkipSubmissionToNextIteration as ex:
                gc3libs.log.info("Submission of job %s delayed", app)
                # Just raise the exception
                raise
            # pylint: disable=broad-except
            except Exception as ex:
                gc3libs.log.info(
                    "Error in submitting job to resource '%s': %s: %s",
                    resource.name, ex.__class__.__name__, str(ex),
                    exc_info=True)
                exs.append(ex)
                continue
            gc3libs.log.info("Successfully submitted %s to: %s",
                             str(app), resource.name)
            job.state = Run.State.SUBMITTED
            job.resource_name = resource.name
            job.info = ("Submitted to '%s'" % (job.resource_name,))
            app.changed = True
            app.submitted()
            # job submitted; return to caller
            return
        # if wet get here, all submissions have failed; call the
        # appropriate handler method if defined
        ex = app.submit_error(exs)
        if isinstance(ex, Exception):
            app.execution.info = ("Submission failed: %s" % str(ex))
            raise ex
        else:
            return

    def __submit_task(self, task, resubmit, targets, **extra_args):
        """Implementation of `submit` on generic `Task` objects."""
        extra_args.setdefault('auto_enable_auth', self.auto_enable_auth)
        task.submit(resubmit, targets, **extra_args)

    def update_job_state(self, *apps, **extra_args):
        """
        Update state of all applications passed in as arguments.

        If keyword argument `update_on_error` is `False` (default),
        then application execution state is not changed in case a
        backend error happens; it is changed to `UNKNOWN` otherwise.

        Note that if state of a job changes, the `Run.state` calls the
        appropriate handler method on the application/task object.

        :raise: `gc3libs.exceptions.InvalidArgument` in case one of
                the passed `Application` or `Task` objects is
                invalid. This can stop updating the state of other
                objects in the argument list.

        :raise: `gc3libs.exceptions.ConfigurationError` if the
                configuration of this `Core` object is invalid or
                otherwise inconsistent (e.g., a resource references a
                non-existing auth section).

        """
        self.__update_application(
            (app for app in apps if isinstance(
                app,
                Application)),
            **extra_args)
        self.__update_task(
            (app for app in apps if not isinstance(
                app,
                Application)),
            **extra_args)

    def __update_application(self, apps, **extra_args):
        """Implementation of `update_job_state` on `Application` objects."""
        update_on_error = extra_args.get('update_on_error', False)
        # auto_enable_auth = extra_args.get(
        #     'auto_enable_auth', self.auto_enable_auth)

        for app in apps:
            state = app.execution.state
            old_state = state
            gc3libs.log.debug(
                "About to update state of application: %s (currently: %s)",
                app,
                state)
            try:
                if state not in [
                        Run.State.NEW,
                        Run.State.TERMINATING,
                        Run.State.TERMINATED,
                ]:
                    lrms = self.get_backend(app.execution.resource_name)
                    try:
                        state = lrms.update_job_state(app)
                    # pylint: disable=broad-except
                    except Exception as ex:
                        gc3libs.log.debug(
                            "Error getting status of application '%s': %s: %s",
                            app, ex.__class__.__name__, ex, exc_info=True)
                        state = Run.State.UNKNOWN
                        # run error handler if defined
                        ex = app.update_job_state_error(ex)
                        if isinstance(ex, Exception):
                            raise ex
                    if state != old_state:
                        app.changed = True
                        # set log information accordingly
                        if (app.execution.state == Run.State.TERMINATING
                                and app.execution.returncode is not None
                                and app.execution.returncode != 0):
                            # there was some error, try to explain
                            app.execution.info = (
                                "Execution failed on resource: %s" %
                                app.execution.resource_name)
                            signal = app.execution.signal
                            if signal in Run.Signals:
                                app.execution.info = (
                                    "Abnormal termination: %s" % signal)
                            else:
                                if os.WIFSIGNALED(app.execution.returncode):
                                    app.execution.info = (
                                        "Remote job terminated by signal %d" %
                                        signal)
                                else:
                                    app.execution.info = (
                                        "Remote job exited with code %d" %
                                        app.execution.exitcode)

                    if state != Run.State.UNKNOWN or update_on_error:
                        app.execution.state = state

            except (gc3libs.exceptions.InvalidArgument,
                    gc3libs.exceptions.ConfigurationError,
                    gc3libs.exceptions.UnrecoverableAuthError,
                    gc3libs.exceptions.FatalError):
                # Unrecoverable; no sense in continuing --
                # pass immediately on to client code and let
                # it handle this...
                raise

            except gc3libs.exceptions.UnknownJob:
                # information about the job is lost, mark it as failed
                app.execution.returncode = (Run.Signals.Lost, -1)
                app.execution.state = Run.State.TERMINATED
                app.changed = True
                continue

            except gc3libs.exceptions.InvalidResourceName:
                # could be the corresponding LRMS has been removed
                # because of an unrecoverable error mark application
                # as state UNKNOWN
                gc3libs.log.warning(
                    "Cannot access computational resource '%s',"
                    " marking task '%s' as UNKNOWN.",
                    app.execution.resource_name, app)
                app.execution.state = Run.State.TERMINATED
                app.changed = True
                continue

            # This catch-all clause is needed otherwise the loop stops
            # at the first erroneous iteration
            #
            # pylint: disable=broad-except
            except Exception as ex:
                if gc3libs.error_ignored(
                        # context:
                        # - module
                        'core',
                        # - class
                        'Core',
                        # - method
                        'update_job_state',
                        # - actual error class
                        ex.__class__.__name__,
                        # - additional keywords
                        'update',
                ):
                    gc3libs.log.warning(
                        "Ignored error in Core.update_job_state(): %s", ex)
                    # print again with traceback at a higher log level
                    gc3libs.log.debug(
                        "(Original traceback follows.)", exc_info=True)
                    continue
                else:
                    # propagate generic exceptions for debugging purposes
                    raise

    # pylint: disable=no-self-use
    def __update_task(self, tasks, **extra_args):
        """Implementation of `update_job_state` on generic `Task` objects."""
        for task in tasks:
            assert isinstance(
                task, Task), "Core.update_job_state: passed an argument" \
                " which is not a `Task` instance."
            task.update_state()

    def fetch_output(self, app, download_dir=None,
                     overwrite=False, changed_only=True, **extra_args):
        """
        Retrieve output into local directory `app.output_dir`.

        If the task is not expected to produce any output (i.e.,
        `app.would_output == False`) then the only effect of this is
        to advance the state of ``TERMINATING`` tasks to
        ``TERMINATED``.

        Optional argument `download_dir` overrides the download location.

        The download directory is created if it does not exist.  If it
        already exists, and the optional argument `overwrite` is
        ``False`` (default), it is renamed with a `.NUMBER` suffix and
        a new empty one is created in its place.  Otherwise, if
        'overwrite` is ``True``, files are downloaded over the ones
        already present; in this case, the `changed_only` argument
        controls which files are overwritten:

        - if `changed_only` is ``True`` (default), then only files for
          which the source has a different size or has been modified
          more recently than the destination are copied;

        - if `changed_only` is ``False``, then *all* files in `source`
          will be copied into `destination`, unconditionally.

        Source files that do not exist at `destination` will be
        copied, independently of the `overwrite` and `changed_only`
        settings.

        If the task is in TERMINATING state, the state is changed to
        `TERMINATED`, attribute `output_dir`:attr: is set to the
        absolute path to the directory where files were downloaded,
        and the `terminated` transition method is called on the `app`
        object.

        Task output cannot be retrieved when `app.execution` is in one
        of the states `NEW` or `SUBMITTED`; an
        `OutputNotAvailableError` exception is thrown in these cases.

        :raise: `gc3libs.exceptions.OutputNotAvailableError` if no
                output can be fetched from the remote job (e.g., the
                Application/Task object is in `NEW` or `SUBMITTED`
                state, indicating the remote job has not started
                running).
        """
        assert isinstance(
            app, Task), "Core.fetch_output: passed an `app` argument " \
            "which is not a `Task` instance."
        if isinstance(app, Application):
            self.__fetch_output_application(
                app, download_dir, overwrite, changed_only, **extra_args)
        else:
            # generic `Task` object
            self.__fetch_output_task(
                app, download_dir, overwrite, changed_only, **extra_args)

    def __fetch_output_application(
            self, app, download_dir, overwrite, changed_only, **extra_args):
        """Implementation of `fetch_output` on `Application` objects."""
        job = app.execution
        if job.state in [Run.State.NEW, Run.State.SUBMITTED]:
            raise gc3libs.exceptions.OutputNotAvailableError(
                "Output not available: '%s' currently in state '%s'"
                % (app, app.execution.state))

        # auto_enable_auth = extra_args.get(
        #     'auto_enable_auth', self.auto_enable_auth)

        # determine download directory
        #
        # pylint: disable=protected-access
        download_dir = app._get_download_dir(download_dir)

        if download_dir is not None:
            # Prepare/Clean download dir
            try:
                if overwrite:
                    if not os.path.exists(download_dir):
                        os.makedirs(download_dir)
                else:
                    utils.mkdir_with_backup(download_dir)
            except Exception as ex:
                gc3libs.log.error(
                    "Failed creating download directory '%s': %s: %s",
                    download_dir,
                    ex.__class__.__name__,
                    str(ex))
                raise

            # download job output
            try:
                lrms = self.get_backend(job.resource_name)
                lrms.get_results(app, download_dir, overwrite, changed_only)
                # clear previous data staging errors
                if job.signal == Run.Signals.DataStagingFailure:
                    job.signal = 0
            except gc3libs.exceptions.InvalidResourceName as err:
                ex = app.fetch_output_error(err)
                if isinstance(ex, Exception):
                    job.info = ("No output could be retrieved: %s" % (ex,))
                    raise ex
                else:
                    return
            except gc3libs.exceptions.RecoverableDataStagingError as rex:
                job.info = ("Temporary failure when retrieving results: %s."
                            " Ignoring error, try again." % str(rex))
                return
            except gc3libs.exceptions.UnrecoverableDataStagingError as ex:
                # pylint: disable=redefined-variable-type
                job.signal = Run.Signals.DataStagingFailure
                ex = app.fetch_output_error(ex)
                if isinstance(ex, Exception):
                    job.info = ("No output could be retrieved: %s" % str(ex))
                    raise ex
            # pylint: disable=broad-except
            except Exception as ex:
                ex = app.fetch_output_error(ex)
                if isinstance(ex, Exception):
                    raise ex

            # successfully downloaded results
            gc3libs.log.debug(
                "Downloaded output of '%s' (which is in state %s)",
                app, job.state)

            app.output_dir = os.path.abspath(download_dir)
            app.changed = True

            if job.state == Run.State.TERMINATING:
                gc3libs.log.debug("Final output of '%s' retrieved", app)

        return Task.fetch_output(app, download_dir)

    def __fetch_output_task(
            self, task, download_dir, overwrite, changed_only, **extra_args):
        """Implementation of `fetch_output` on generic `Task` objects."""
        return task.fetch_output(
            download_dir, overwrite, changed_only, **extra_args)

    def get_resources(self, **extra_args):
        """
        Return list of resources configured into this `Core` instance.
        """
        return [lrms for lrms in self.resources.values()]

    def kill(self, app, **extra_args):
        """
        Terminate a job.

        Terminating a job in RUNNING, SUBMITTED, or STOPPED state
        entails canceling the job with the remote execution system;
        terminating a job in the NEW or TERMINATED state is a no-op.
        """
        assert isinstance(
            app, Task), "Core.kill: passed an `app` argument which is not"\
            " a `Task` instance."
        if isinstance(app, Application):
            self.__kill_application(app, **extra_args)
        else:
            self.__kill_task(app, **extra_args)

    def __kill_application(self, app, **extra_args):
        """Implementation of `kill` on `Application` objects."""
        job = app.execution
        # auto_enable_auth = extra_args.get(
        #     'auto_enable_auth', self.auto_enable_auth)
        try:
            lrms = self.get_backend(job.resource_name)
            lrms.cancel_job(app)
        except AttributeError:
            # A job in state NEW does not have a `resource_name`
            # attribute.
            if job.state != Run.State.NEW:
                raise
        except gc3libs.exceptions.InvalidResourceName:
            gc3libs.log.warning(
                "Cannot access computational resource '%s',"
                " but marking task '%s' as TERMINATED anyway.",
                app.execution.resource_name, app)
        gc3libs.log.debug(
            "Setting task '%s' status to TERMINATED"
            " and returncode to SIGCANCEL", app)
        app.changed = True
        # setting the state runs the state-transition handlers,
        # which may raise an error -- ignore them, but log nonetheless
        try:
            job.state = Run.State.TERMINATED
        # pylint: disable=broad-except
        except Exception as ex:
            if gc3libs.error_ignored(
                    # context:
                    # - module
                    'core',
                    # - class
                    'Core',
                    # - method
                    'kill',
                    # - actual error class
                    ex.__class__.__name__,
                    # - additional keywords
                    'state',
                    job.state,
                    'TERMINATED',
            ):
                gc3libs.log.info("Ignoring error in state transition"
                                 " since task is being killed: %s", ex)
            else:
                # propagate exception to caller
                raise
        job.signal = Run.Signals.Cancelled
        job.history.append("Cancelled")

    def __kill_task(self, task, **extra_args):
        extra_args.setdefault('auto_enable_auth', self.auto_enable_auth)
        task.kill(**extra_args)

    def peek(self, app, what='stdout', offset=0, size=None, **extra_args):
        """
        Download `size` bytes (at `offset` bytes from the start) from
        the remote job standard output or error stream, and write them
        into a local file.  Return file-like object from which the
        downloaded contents can be read.

        If `size` is `None` (default), then snarf all available
        contents of the remote stream from `offset` unto the end.

        The only allowed values for the `what` arguments are the
        strings `'stdout'` and `'stderr'`, indicating that the
        relevant section of the job's standard output resp. standard
        error should be downloaded.
        """
        assert isinstance(
            app, Task), "Core.peek: passed an `app` argument which is" \
            " not a `Task` instance."
        if isinstance(app, Application):
            return self.__peek_application(
                app, what, offset, size, **extra_args)
        else:
            return self.__peek_task(app, what, offset, size, **extra_args)

    def __peek_application(self, app, what, offset, size, **extra_args):
        """Implementation of `peek` on `Application` objects."""
        if what == 'stdout':
            remote_filename = app.stdout
        elif what == 'stderr':
            remote_filename = app.stderr
        else:
            raise gc3libs.exceptions.Error(
                "File name requested to `Core.peek` must be"
                " 'stdout' or 'stderr', not '%s'" % what)

        # Check if local data available
        job = app.execution
        if job.state == Run.State.TERMINATED:
            # FIXME: local data could be stale!!
            filename = os.path.join(app.output_dir, remote_filename)
            local_file = open(filename, 'r')
        else:
            # Get authN
            # auto_enable_auth = extra_args.get(
            #     'auto_enable_auth', self.auto_enable_auth)
            lrms = self.get_backend(job.resource_name)
            local_file = tempfile.NamedTemporaryFile(
                suffix='.tmp', prefix='gc3libs.', mode='w+t')
            lrms.peek(app, remote_filename, local_file, offset, size)
            local_file.flush()
            local_file.seek(0)

        return local_file

    def __peek_task(self, task, what, offset, size, **extra_args):
        """Implementation of `peek` on generic `Task` objects."""
        return task.peek(what, offset, size, **extra_args)

    def update_resources(self, resources=all, **extra_args):
        """
        Update the state of a given set of resources.

        Each resource object in the returned list will have its `updated`
        attribute set to `True` if the update operation succeeded, or `False`
        if it failed.

        Optional argument `resources` should be a subset of the
        resources configured in this `Core` instance (the actual
        `Lrms`:class: objects, not the resource names).  By default,
        all configured resources are updated.
        """
        if resources is all:
            resources = list(self.resources.values())
        for lrms in self.resources.values():
            try:
                if not lrms.enabled:
                    continue
                # auto_enable_auth = extra_args.get(
                #     'auto_enable_auth', self.auto_enable_auth)
                lrms.get_resource_status()
                lrms.updated = True
            except gc3libs.exceptions.UnrecoverableError as err:
                # disable resource -- there's no point in
                # trying it again at a later stage
                lrms.enabled = False
                lrms.updated = False
                gc3libs.log.error(
                    "Unrecoverable error updating status"
                    " of resource '%s': %s."
                    " Disabling resource.",
                    lrms.name, err)
                gc3libs.log.warning(
                    "Resource %s will be ignored from now on.",
                    lrms.name)
                gc3libs.log.debug(
                    "Got error '%s' in updating resource '%s';"
                    " printing full traceback.",
                    err.__class__.__name__, lrms.name,
                    exc_info=True)
            # pylint: disable=broad-except
            except Exception as err:
                gc3libs.log.error(
                    "Ignoring error updating resource '%s': %s.",
                    lrms.name, err)
                gc3libs.log.debug(
                    "Got error '%s' in updating resource '%s';"
                    " printing full traceback.",
                    err.__class__.__name__, lrms.name,
                    exc_info=True)
                lrms.updated = False

    def close(self):
        """
        Used to invoke explicitly the destructor on objects
        e.g. LRMS
        """
        for lrms in self.resources.values():
            lrms.close()

    # compatibility with the `Engine` interface

    def add(self, task):
        """
        This method is here just to allow `Core` and `Engine` objects
        to be used interchangeably.  It's effectively a no-op, as it makes
        no sense in the synchronous/blocking semantics implemented by `Core`.
        """
        pass

    def remove(self, task):
        """
        This method is here just to allow `Core` and `Engine` objects
        to be used interchangeably.  It's effectively a no-op, as it makes
        no sense in the synchronous/blocking semantics implemented by `Core`.
        """
        pass


class Scheduler(object):
    """
    Instances of the `Scheduler` class are used in
    `Engine.progress`:meth: to determine what tasks (among those in
    `Run.State.NEW` state) are to be submitted.

    A `Scheduler` object must implement *both* the context_ protocol
    *and* the iterator_ protocol.

    .. _context:  http://goo.gl/SvWWyw
    .. _iterator: http://goo.gl/ue2zje

    The way a `Scheduler` instance is actually used within `Engine` is
    as follows:

    0. A `Scheduler` instance is created, passing it two arguments: a
       list of tasks in ``NEW`` state, and a dictionary of configured
       resources (keys are resource names, values are actual resource
       objects).

    1. When a new submission cycle starts, the `__enter__`:meth:
       method is called.

    2. The `Engine` iterates by repeatedly calling the `next`:meth:
       method to receive tasks to be submitted.  The `send`:meth: and
       `throw`:meth: methods are used to notify the scheduler of the
       outcome of the submission attempt.

    3. When the submission cycle ends, the `__exit__`:meth: method is called.

    The `Scheduler.schedule` generator is the heart of the submission
    process and has basically complete control over it.  It is
    initialized with the list of tasks in ``NEW`` state, and the list
    of configured resources.  The `next`:meth: method should yield
    pairs *(task index, resource name)*, where the *task index* is the
    position of the task to be submitted next in the given list, and
    --similarly-- the *resource name* is the name of the resource to
    which the task should be submitted.

    For each pair yielded, submission of that task to the selected
    resource is attempted; the state of the task object after
    submission is sent back (via the `send`:meth: method) to the
    `Scheduler` instance; if an exception is raised, that exception is
    thrown (via the `throw`:meth: method) into the scheduler object
    instead.  Submission stops when the `next()` call raises a
    `StopIteration` exception.
    """

    def __init__(self, tasks, resources):
        self.tasks = tasks
        self.resources = resources

    def __enter__(self):
        """Called at the start of a scheduling cycle.

        Implementation of this method should follow Python's context_
        protocol; in particular, this method must return a reference
        to a valid context object.

        By default, just returns a reference to ``self``.

        """
        return self

    # pylint: disable=missing-docstring
    def __next__(self):
        raise NotImplementedError(
            "Method `next` of class `%s` has not been implemented."
            % self.__class__.__name__)

    # pylint: disable=missing-docstring
    def send(self, result):
        raise NotImplementedError(
            "Method `send` of class `%s` has not been implemented."
            % self.__class__.__name__)

    # pylint: disable=missing-docstring
    def throw(self, *excinfo):
        raise NotImplementedError(
            "Method `throw` of class `%s` has not been implemented."
            % self.__class__.__name__)

    def __exit__(self, *excinfo):
        """Called at the end of a scheduling cycle, when no more submissions
        will be performed by the `Engine`.

        Implementation of this method should follow Python's context_ protocol.

        By default, does nothing.

        :param tuple excinfo:
          This is either:

            - a triple *(exception class, exception value, traceback)*
              like the one returned from the standard `sys.exc_info`
              function, when an unhandled error occurred during the
              submission cycle, or
            - the empty tuple if execution of the submission cycle
              terminated normally without exceptions.

        """
        pass


class scheduler(object):  # pylint: disable=invalid-name

    """
    Decorate a generator function for use as a `Scheduler`:class: object.
    """
    __slots__ = ['_fn', '_gen']

    def __init__(self, fn):
        self._fn = fn

    def __call__(self, *args, **kwargs):
        # pylint: disable=attribute-defined-outside-init
        self._gen = self._fn(*args, **kwargs)
        return self

    #
    # proxy generator protocol methods
    #

    def __iter__(self):
        return self

    # pylint: disable=missing-docstring
    def __next__(self):
        return next(self._gen)

    # pylint: disable=missing-docstring
    def send(self, value):
        return self._gen.send(value)

    # pylint: disable=missing-docstring
    def throw(self, *excinfo):
        return self._gen.throw(*excinfo)

    # pylint: disable=missing-docstring
    def close(self):
        self._gen.close()

    #
    # add context protocol methods
    #

    def __enter__(self):
        return self

    def __exit__(self, *excinfo):
        self.close()


@scheduler
def first_come_first_serve(task_queue, resources, matchmaker=MatchMaker()):
    """
    First-come first-serve scheduling policy.

    Tasks are submitted to resources in the order they appear in the
    `task_queue` queue.  Each task is submitted to resources according to
    the order they are sorted by `Application.rank_resources` (if that
    method exists).

    This is the default scheduling policy in GC3Pie's `Engine`:class.
    """
    assert resources, "No execution resources available!"
    # make a copy of the `resources` argument, so we can modify it
    # when e.g. disabling resources that are full
    resources = list(resources)
    total = len(task_queue)
    for done in range(total):
        task = task_queue.get()
        # keep only compatible resources
        compatible_resources = matchmaker.filter(task, resources)
        if not compatible_resources:
            gc3libs.log.warning(
                "No compatible resources for task '%s'"
                " - cannot submit it", task)
            task_queue.put(task)
            continue
        # sort them according to the Task's preference
        targets = matchmaker.rank(task, compatible_resources)
        # now try submission of the task to each resource until one succeeds
        for target in targets:
            try:
                yield (task, target.name)
                # submission successful, continue with next task
                break
            except (gc3libs.exceptions.ResourceNotReady,
                    gc3libs.exceptions.MaximumCapacityReached) as exc:
                # this is not a real error: the resource is adapting
                # for the task and will actually accept it sometime in
                # the future, so disable resource and try next one
                gc3libs.log.debug(
                    "Disabling resource `%s` for this scheduling cycle: %s",
                    target.name, exc)
                resources.remove(target)
                continue
            # pylint: disable=broad-except
            except Exception as err:
                # note error condition but continue with next target resource
                gc3libs.log.debug(
                    "Scheduler ignored error in submitting task '%s': %s: %s",
                    task, err.__class__.__name__, err, exc_info=True)
                continue
        else:
            # unsuccessful submission, push task back into queue
            task_queue.put(task)
        if not resources:
            gc3libs.log.debug(
                "No more resources available,"
                " aborting scheduling cycle with %d tasks remaining.",
                total - done)
            return


class Engine(object):  # pylint: disable=too-many-instance-attributes
    """
    Manage a collection of tasks, until a terminal state is reached.
    Specifically:

    * tasks in `NEW` state are submitted;

    * the state of tasks in `SUBMITTED`, `RUNNING` or `STOPPED` state
      is updated;

    * when a task reaches `TERMINATED` state, its output is downloaded.

    The behavior of `Engine` instances can be further customized by
    setting the following instance attributes:

    `can_submit`
      Boolean value: if `False`, no task will be submitted.

    `can_retrieve`
      Boolean value: if `False`, no output will ever be retrieved.

    `max_in_flight`
      If >0, limit the number of tasks in `SUBMITTED` or `RUNNING`
      state: if the number of tasks in `SUBMITTED`, `RUNNING` or
      `STOPPED` state is greater than `max_in_flight`, then no new
      submissions will be attempted.

    `max_submitted`
      If >0, limit the number of tasks in `SUBMITTED` state: if the
      number of tasks in `SUBMITTED` state is greater than
      `max_submitted`, then no new submissions will be attempted.

    `output_dir`
      Base directory for job output; if not `None`, each task's
      results will be downloaded in a subdirectory named after the
      task's `permanent_id`.

    `scheduler`
      A factory function for creating objects that conform to the
      `Scheduler` interface to control task submission; see the
      `Scheduler`:class: documentation for details.  The default value
      implements a first-come first-serve algorithm: tasks are
      submitted in the order they have been added to the `Engine`.

    `retrieve_running`
      If ``True``, snapshot output from RUNNING jobs at every
      invocation of `progress`:meth:

    `retrieve_overwrites`
      If ``True``, overwrite files in the output directory of any job
      (as opposed to moving destination away and downloading a fresh
      copy). See `Core.fetch_output`:meth: for details.

    `retrieve_changed_only`
      If both this and `overwrite` are ``True``, then only changed
      files are downloaded. See `Core.fetch_output`:meth: for details.

    `forget_terminated`
      When ``True``, `Engine.remove`:meth: is automatically called
      on tasks when their state turns to ``TERMINATED``.

      .. warning::

        For historical reasons, the default for this option is
        ``False`` but this can (and should!) be changed in future
        releases.

    Any of the above can also be set by passing a keyword argument to
    the constructor (assume ``g`` is a `Core`:class: instance)::

    | >>> e = Engine(g, can_submit=False)
    | >>> e.can_submit
    | False
    """

    def __init__(self, controller, tasks=[], store=None,
                 can_submit=True, can_retrieve=True,
                 max_in_flight=0, max_submitted=0,
                 output_dir=None,
                 scheduler=first_come_first_serve,  # pylint: disable=redefined-outer-name
                 retrieve_running=False,
                 retrieve_overwrites=False,
                 retrieve_changed_only=True,
                 forget_terminated=False):
        """
        Create a new `Engine` instance.  Arguments are as follows:

        :param controller:
          A `gc3libs.Core` instance, that will be used to operate on
          tasks.  This is the only required argument.

        :param list apps:
          Initial list of tasks to be managed by this Engine.  Tasks can
          be later added and removed with the `add` and `remove`
          methods (which see).  Defaults to the empty list.

        :param store:
          An instance of `gc3libs.persistence.Store`, or `None`.  If
          not `None`, it will be used to persist tasks after each
          iteration; by default no store is used so no task state is
          persisted.

        :param can_submit:
        :param can_retrieve:
        :param max_in_flight:
        :param max_submitted:
        :param output_dir:
        :param scheduler:
        :param bool retrieve_running:
        :param bool retrieve_overwrites:
        :param bool retrieve_changed_only:
          Optional keyword arguments; see `Engine`:class: for a description.

        """
        # internal-use attributes
        self._managed = self._TaskQueueManager(self)
        self._core = controller
        self._store = store
        self._tasks_by_id = {}

        # public attributes
        self.can_submit = can_submit
        self.can_retrieve = can_retrieve
        self.max_in_flight = max_in_flight
        self.max_submitted = max_submitted
        self.output_dir = output_dir
        self.scheduler = scheduler
        self.retrieve_running = retrieve_running
        self.retrieve_overwrites = retrieve_overwrites
        self.retrieve_changed_only = retrieve_changed_only
        self.forget_terminated = forget_terminated

        # init counters/statistics
        self._counts = self._Counters(self)
        # always gather statistics about `Task` instances, as `Task`
        # is the root of the GC3Pie task hierarchy
        self._counts.init_for(Task)
        # but also count `Application` instances to enforce submission
        # and running limits (only real tasks that consume compute
        # resources count for those -- "policy" tasks like
        # `TaskCollection` should not)
        self._counts.init_for(Application)
        TaskStateChange.connect(self._on_state_change)

        # Engine fully initialized, add all tasks
        for task in tasks:
            self.add(task)

    def _on_state_change(self, task, from_state, to_state):
        if task in self._managed:
            #gc3libs.log.debug("Task %s transitioned from %s to %s ...", task, from_state, to_state)
            self._counts.transitioned(task, from_state, to_state)

    class TaskQueue(object):
        def __init__(self):
            # the actual data structure used for keeping tasks
            self._queue = deque()

        def __iter__(self):
            return iter(self._queue)

        def __len__(self):
            return len(self._queue)

        def add(self, task):
            """
            Add task in front of queue.

            Does *not* check if already present.
            """
            self._queue.appendleft(task)

        def get(self):
            """
            Pop the first task from the front of queue.
            """
            task = self._queue.popleft()
            return task

        def put(self, task):
            """
            Add task to back of queue.

            Does *not* check if already present.
            """
            self._queue.append(task)

        def remove(self, task):
            """
            Remove the given task from the queue.
            """
            # FIXME: this relies on `collections.deque` which supports
            # removal of an element by value; Python's `queue.Queue`
            # only provides put/get operations and would make this
            # difficult to do.  We would need to find a different
            # solution or data structure before we can generalize the
            # Engine to span multiple processes...
            self._queue.remove(task)


    class _TaskQueueManager(object):

        def __init__(self, engine):
            # action-based queues
            self.to_kill = engine.TaskQueue()
            self.to_update = engine.TaskQueue()
            self.to_submit = engine.TaskQueue()
            self.to_fetch_output = engine.TaskQueue()
            self.to_cleanup = engine.TaskQueue()
            self.done = engine.TaskQueue()

            # map action names to queues
            self._actions = {
                'kill':         self.to_kill,
                'update':       self.to_update,
                'submit':       self.to_submit,
                'fetch_output': self.to_fetch_output,
                'cleanup':      self.to_cleanup,
                'done':         self.done,
            }

            # map task ID to action
            self._index = {}

        def __contains__(self, task):
            """
            Return ``True`` if `task` is present in some queue.
            """
            return id(task) in self._index

        def get_queue(self, task, _override_state=None):
            """
            Return the "queue" object to which `task` should be added or removed.
            """
            state = _override_state or task.execution.state
            if Run.State.NEW == state:
                return self.to_submit
            elif state in [
                    Run.State.RUNNING,
                    Run.State.STOPPED,
                    Run.State.SUBMITTED,
                    Run.State.UNKNOWN,
            ]:
                return self.to_update
            elif Run.State.TERMINATING == state:
                return self.to_fetch_output
            elif Run.State.TERMINATED == state:
                # FIXME: we should never add here when `self.forget_terminated == True`
                return self.done
            else:
                raise AssertionError(
                    "Unhandled state '%s' in gc3libs.core.Engine." % state)

        def add(self, task, action=None):
            """
            Append a task to the back of the appropriate action queue.
            """
            queue = (self._actions[action] if action else self.get_queue(task))
            queue.put(task)
            self._index[id(task)] = queue

        def remove(self, task):
            """
            Delete the given task from the queue it is in.
            After calling this method, `task in this` will be ``False``.
            """
            queue = self._index.pop(id(task))
            queue.remove(task)

        def requeue(self, task, action=None):
            """
            Move an already-managed task to the back of the queue
            that is appropriate for its state.
            """
            try:
                queue = self._index[id(task)]
                queue.remove(task)
            except (KeyError, ValueError):
                pass
            self.add(task, action)


    class _Counters(object):
        """
        Keep count of how many tasks of a class are in a given state.
        """
        def __init__(self, engine):
            self._engine = engine
            self.totals = {}

        def init_for(self, cls):
            """
            Initialize counters for tasks of class `cls`.

            All statistics are initially computed starting from the
            current collection of tasks managed by the parent `Engine`
            instance; they will be kept up-to-date during task
            addition/removal/progress.
            """
            # FIXME: rewrite using `collections.Counter` when we drop
            # support for Py 2.6?
            counter = self.totals[cls] = defaultdict(int)
            for task in self._engine.iter_tasks(cls):
                counter['total'] += 1
                state = task.execution.state
                counter[state] += 1
                if state == Run.State.TERMINATED:
                    if task.execution.returncode == 0:
                        counter['ok'] += 1
                    else:
                        counter['failed'] += 1
            return counter

        def _update(self, task, increment):
            """
            Update the counts relative to `task`'s state by `increment`.
            """
            state = task.execution.state
            stats_to_update = ['total', state]
            if state == 'TERMINATED':
                if task.execution.returncode == 0:
                    stats_to_update.append('ok')
                else:
                    stats_to_update.append('failed')
            for cls in self.totals:
                if isinstance(task, cls):
                    for stat in stats_to_update:
                        self.totals[cls][stat] += increment

        def add(self, task):
            """
            Adjust totals to include `task`.
            """
            self._update(task, +1)

        def remove(self, task):
            """
            Adjust totals following the removal of `task`.
            """
            self._update(task, -1)

        def transitioned(self, task, from_state, to_state):
            """
            Update the counts, following a `TaskStateChange` event.

            The counters relative to *from_state* are decremented by
            unit, and correspondingly the counters for *to_state* are
            incremented.

            This is functionally equivalent to, but more efficient than::

              self._update(task, from_state, -1)
              self._update(task, to_state, +1)
            """
            stats_to_increment = [to_state]
            if to_state == 'TERMINATED':
                if task.execution.returncode == 0:
                    stats_to_increment.append('ok')
                else:
                    stats_to_increment.append('failed')
                # update counts if exit code is later changed
                TermStatusChange.connect(self._on_termstatus_change, sender=task)
            for cls in self.totals:
                if isinstance(task, cls):
                    self.totals[cls][from_state] -= 1
                    for stat in stats_to_increment:
                        self.totals[cls][stat] += 1

        def _on_termstatus_change(self, task, from_returncode, to_returncode):
            # shortcut
            if from_returncode == to_returncode:
                return
            # decrement the "from" status
            target = 'ok' if from_returncode == 0 else 'failed'
            for cls in self.totals:
                if isinstance(task, cls):
                    self.totals[cls][target] -= 1
            # remove handler if task was resubmitted
            if task.execution.state != 'TERMINATED':
                TermStatusChange.disconnect(self._on_termstatus_change, sender=task)
            # increment the "to" status
            target = 'ok' if to_returncode == 0 else 'failed'
            for cls in self.totals:
                if isinstance(task, cls):
                    self.totals[cls][target] += 1


    def add(self, task):
        """
        Add `task` to the list of tasks managed by this Engine.
        Adding a task that has already been added to this `Engine`
        instance results in a no-op.
        """
        if task not in self._managed:
            self._managed.add(task)
            self._counts.add(task)
            if self._store:
                try:
                    self._tasks_by_id[task.persistent_id] = task
                except AttributeError:
                    gc3libs.log.debug(
                        "Engine %s: Added task %s with no persistent ID!",
                        self, task)
            task.attach(self)


    def remove(self, task):
        """
        Remove a `task` from the list of tasks managed by this Engine.

        Removing a task that is not managed (i.e., already removed or
        never added) is a no-op.
        """
        if task not in self._managed:
            return
        self._managed.remove(task)
        self._counts.remove(task)
        if self._store:
            try:
                del self._tasks_by_id[task.persistent_id]
            except KeyError:
                # already removed
                pass
            except AttributeError:
                gc3libs.log.debug(
                    "Engine %s: Cannot remove task %s: has no persistent ID!",
                    self, task)
        task.detach()


    def find_task_by_id(self, task_id):
        """
        Return the task with the given persistent ID added to this
        `Engine` instance.  If no task has that ID, raise a `KeyError`.
        """
        return self._tasks_by_id[task_id]


    def iter_tasks(self, only_cls=None):
        """
        Iterate over tasks managed by the Engine.

        If argument `only_cls` is ``None`` (default), then iterate over
        *all* tasks managed by this Engine.  Otherwise, only return
        tasks which are instances of a (sub)class `only_cls`.
        """
        if only_cls is None:
            select = self.__iter_all
        else:
            select = self.__iter_only
        return itertools.chain(
            select(self._managed.to_submit, only_cls),
            select(self._managed.to_update, only_cls),
            select(self._managed.to_kill, only_cls),
            select(self._managed.to_fetch_output, only_cls),
            select(self._managed.to_cleanup, only_cls),
            select(self._managed.done, only_cls),
        )

    # helper methods for `iter_tasks`; they are created as
    # "staticmethod"s instead of `lambda`-functions to save creating a
    # closure for each invocation of `iter_tasks`
    @staticmethod
    def __iter_all(queue, _):
        return iter(queue)

    @staticmethod
    def __iter_only(queue, cls):
        return filter(
            (lambda task: isinstance(task, cls)), iter(queue))


    def init_counts_for(self, cls):
        """
        Initialize counters for tasks of class `cls`.

        All statistics are initially computed starting from the current
        collection of tasks managed by this `Engine` instance; they will
        be kept up-to-date during task addition/removal/progress.

        .. warning::

          In a future release, the `Engine` might forget about task
          objects in ``TERMINATED`` state.  Therefore, `init_counts_for`
          should be called before any tasks reaches ``TERMINATED``
          state, or the counts for ``TERMINATED``, ``ok``, and
          ``failed`` jobs will be incorrectly initialized to 0.
        """
        counter = self._counts.init_for(cls)
        if Run.State.TERMINATED in counter and counter[Run.State.TERMINATED] > 0:
            warn("The Engine class will forget TERMINATED tasks in the near future."
                 "In order to get correct results, `init_counts_for`"
                 " should be called before any task reaches TERMINATED state",
                 FutureWarning)


    def progress(self):
        """
        Update state of all registered tasks and take appropriate action.
        Specifically:

          * tasks in `NEW` state are submitted;

          * the state of tasks in `SUBMITTED`, `RUNNING`, `STOPPED` or
            `UNKNOWN` state is updated;

          * when a task reaches `TERMINATING` state, its output is downloaded.

          * tasks in `TERMINATED` status are simply ignored.

        The `max_in_flight` and `max_submitted` limits (if >0) are
        taken into account when attempting submission of tasks.
        """
        gc3libs.log.debug("Engine.progress(): starting.")

        # pylint: disable=redefined-variable-type
        if self.max_in_flight > 0:
            limit_in_flight = self.max_in_flight
        else:
            limit_in_flight = utils.PlusInfinity()
        if self.max_submitted > 0:
            limit_submitted = self.max_submitted
        else:
            limit_submitted = utils.PlusInfinity()

        # if no resources are enabled, there's no point in running
        # this further
        nr_enabled_resources = sum(int(rsc.enabled)
                                   for rsc in self._core.resources.values())
        if nr_enabled_resources == 0:
            raise gc3libs.exceptions.NoResources(
                "No resources available for running jobs.")

        # execute kills and update count of submitted/in-flight tasks
        queue = self._managed.to_kill
        if queue:
            gc3libs.log.debug("Engine %s about to kill jobs ...", self)
        for _ in range(len(queue)):
            task = queue.get()
            old_state = task.execution.state

            try:
                self._core.kill(task)
                if self._store and task.changed:
                    self._store.save(task)
            # pylint: disable=broad-except
            except Exception as err:
                self.__ignore_or_raise(
                    err, "killing", task,
                    # context:
                    # - module
                    'core',
                    # - class
                    'Engine',
                    # - method
                    'progress',
                    # - actual error class
                    err.__class__.__name__,
                    # - additional keywords
                    'kill'
                )

            self._managed.requeue(task)

        # update status of tasks before launching new ones
        queue = self._managed.to_update
        if queue:
            gc3libs.log.debug(
                "Engine %s about to update status of in-flight tasks ...",
                self)
        for _ in range(len(queue)):
            task = queue.get()

            # ensure pre-condition on state is met
            old_state = task.execution.state
            if old_state not in [
                    Run.State.RUNNING,
                    Run.State.STOPPED,
                    Run.State.SUBMITTED,
                    Run.State.UNKNOWN,
            ]:
                # task changed state outside of the Engine, requeue
                self._managed.requeue(task)
                continue

            try:
                self._core.update_job_state(task)
                if self._store and task.changed:
                    self._store.save(task)
            except gc3libs.exceptions.ConfigurationError:
                # Unrecoverable; no sense in continuing -- pass
                # immediately on to client code and let it handle
                # this...
                raise
            # pylint: disable=broad-except
            except Exception as err:
                self.__ignore_or_raise(
                    err, "updating state", task,
                    # context:
                    # - module
                    'core',
                    # - class
                    'Engine',
                    # - method
                    'progress',
                    # - actual error class
                    err.__class__.__name__,
                    # - additional keywords
                    'state',
                    'update',
                )

            # do book-keeping
            state = task.execution.state

            if state in [
                    Run.State.RUNNING,
                    Run.State.STOPPED,
                    Run.State.SUBMITTED,
                    Run.State.UNKNOWN,
            ]:
                queue.put(task)
            else:
                self._managed.requeue(task)

            if self.retrieve_running:
                if (state == Run.State.RUNNING and task.would_output
                        and self.can_retrieve):
                    # try to get output
                    try:
                        self._core.fetch_output(
                            task,
                            overwrite=self.retrieve_overwrites,
                            changed_only=self.retrieve_changed_only)
                    # pylint: disable=broad-except
                    except Exception as err:
                        self.__ignore_or_raise(
                            err, "fetching output", task,
                            # context:
                            # - module
                            'core',
                            # - class
                            'Engine',
                            # - method
                            'progress',
                            # - actual error class
                            err.__class__.__name__,
                            # - additional keywords
                            'RUNNING',
                            'fetch_output',
                        )

        # reckon how many tasks are "live"; we are only interested in
        # tasks that consume real compute resources (i.e.,
        # `Application` instances) and exclude "policy" classes (e.g.,
        # all `TaskCollections`)
        app_counts = self.counts(Application)
        currently_submitted = app_counts['SUBMITTED']
        currently_in_flight = currently_submitted + app_counts['RUNNING']

        # now try to submit NEW tasks
        if (self.can_submit and
                currently_submitted < limit_submitted and
                currently_in_flight < limit_in_flight):
            submit_allowance = min(
                limit_submitted - currently_submitted,
                limit_in_flight - currently_in_flight
            )
            queue = self._managed.to_submit
            if queue:
                gc3libs.log.debug(
                    "Engine %s about to submit NEW tasks ...", self)
            # update state of all enabled resources, to give a chance to
            # all to get a new job; for a complete discussion, see:
            # https://github.com/uzh/gc3pie/issues/485
            self._core.update_resources()
            # now try to submit
            with self.scheduler(queue,
                                list(self._core.resources.values())) as _sched:
                # wrap the original generator object so that `send`
                # and `throw` do not yield a value -- we only get new
                # stuff from the call to the `next` method in the `for
                # ... in sched:` line
                sched = gc3libs.utils.YieldAtNext(_sched)
                for task, resource_name in sched:
                    # enforce Engine limits
                    if submit_allowance <= 0:
                        # we need to put back the task in the queue
                        # here as we won't go call back into scheduler
                        self._managed.to_submit.put(task)
                        break
                    resource = self._core.resources[resource_name]
                    try:
                        self._core.submit(task, targets=[resource])
                        # if we get to this point, we know state is
                        # either SUBMITTED or RUNNING
                        if self._store and task.changed:
                            self._store.save(task)
                        self._managed.to_update.put(task)
                        if isinstance(task, Application):
                            submit_allowance -= 1
                        # notify scheduler
                        sched.send(task.execution.state)
                    # pylint: disable=broad-except
                    except Exception as err1:
                        # record the error in the task's history
                        task.execution.history(
                            "Submission to resource '%s' failed: %s: %s"
                            % (resource.name, err1.__class__.__name__, err1))
                        gc3libs.log.error(
                            "Got error in submitting task '%s',"
                            " informing scheduler: %s: %s",
                            task, err1.__class__.__name__, err1)
                        # inform scheduler and let it handle it
                        try:
                            sched.throw(* sys.exc_info())
                        # pylint: disable=broad-except
                        except Exception as err2:
                            self.__ignore_or_raise(
                                err2, "submitting task", task,
                                # context:
                                # - module
                                'core',
                                # - class
                                'Engine',
                                # - method
                                'progress',
                                # - actual error class
                                err2.__class__.__name__,
                                # - additional keywords
                                'scheduler',
                                'submit',
                            )

        # finally, retrieve output of finished tasks
        if self.can_retrieve:
            queue = self._managed.to_fetch_output
            if queue:
                gc3libs.log.debug(
                    "Engine %s about to retrieve output of TERMINATING tasks ...",
                    self)
            for _ in range(len(queue)):
                task = queue.get()
                # try to get output
                try:
                    self._core.fetch_output(
                        task,
                        overwrite=self.retrieve_overwrites,
                        changed_only=self.retrieve_changed_only)
                except gc3libs.exceptions.UnrecoverableDataStagingError as err:
                    gc3libs.log.error(
                        "Error in fetching output of task '%s',"
                        " will mark it as TERMINATED"
                        " (with error exit code %d): %s: %s",
                        task, posix.EX_IOERR,
                        err.__class__.__name__, err, exc_info=True)
                    task.execution.returncode = (
                        Run.Signals.DataStagingFailure,
                        posix.EX_IOERR)
                    task.execution.state = Run.State.TERMINATED
                    task.changed = True
                # pylint: disable=broad-except
                except Exception as err:
                    self.__ignore_or_raise(
                        err, "fetching output", task,
                        # context:
                        # - module
                        'core',
                        # - class
                        'Engine',
                        # - method
                        'progress',
                        # - actual error class
                        err.__class__.__name__,
                        # - additional keywords
                        'fetch_output',
                    )

                if task.execution.state == Run.State.TERMINATED:
                    self._managed.requeue(task, 'cleanup')
                else:
                    assert task.execution.state == 'TERMINATING'
                    queue.put(task)  # retry next time

        # clean up terminated tasks
        queue = self._managed.to_cleanup
        if queue:
            gc3libs.log.debug(
                "Engine %s about to clean up TERMINATED tasks ...", self)
        for _ in range(len(queue)):
            task = queue.get()
            try:
                self._core.free(task)
                self._managed.requeue(task, 'done')
            # pylint: disable=broad-except
            except Exception as err:
                queue.put(task)  # retry next time
                gc3libs.log.error(
                    "Got error freeing up resources used by task '%s': %s: %s."
                    " (For cloud-based resources, it's possible that the VM"
                    " has been destroyed already.)",
                    task, err.__class__.__name__, err)

            if self._store and task.changed:
                self._store.save(task)

        # now remove all terminated tasks
        if self.forget_terminated:
            queue = self._managed.done
            if queue:
                gc3libs.log.debug(
                    "Engine %s now dropping TERMINATED tasks ...", self)
            # the whole purpose of this loop is to call
            # `self.remove()` on each item in the queue, so we cannot
            # use the dequeue/requeue mechanism used in other loops,
            # otherwise `self.remove()` fails since the task is not in
            # `self._managed.done` ...
            for task in list(queue):
                try:
                    self.remove(task)
                    gc3libs.log.debug(
                        "Dropped TERMINATED task %s", task)
                except Exception as err:  # pylint: disable=broad-except
                    gc3libs.log.debug(
                        "Could not forget TERMINATED task '%s': %s: %s",
                        task, err.__class__.__name__, err)

        gc3libs.log.debug("Engine.progress(): done.")


    def __ignore_or_raise(self, err, action, task, *ctx):
        if gc3libs.error_ignored(*ctx):
            gc3libs.log.debug(
                "Ignored error in %s of task '%s': %s: %s",
                action, task, err.__class__.__name__, err)
            gc3libs.log.debug(
                "(Original traceback follows.)",
                exc_info=True)
        else:
            # propagate exceptions for debugging purposes
            raise


    def redo(self, task, *args, **kwargs):
        """
        Reset task's state to NEW so that it will be re-run.

        Any additional arguments will be forwarded to the task's own
        `.redo()` method; this is useful, e.g., to perform partial
        re-runs of `SequentialTaskCollection` instances.
        """
        self.remove(task)
        task.redo(*args, **kwargs)
        self.add(task)


    @property
    def resources(self):
        """
        Get dict of configured resources.

        This mapping object has configured resource names as keys, and the
        actual `gc3libs.backends.LRMS` instances as values. Note that only
        resources whose ``.enabled`` attribute evaluates to ``True`` will be
        considered for scheduling.

        This is just a reference to the ``.resources`` attribute of the
        underlying core object; see `Core.resources` for more information.
        """
        return self._core.resources


    # FIXME: rewrite using `collections.Counter` when we drop support for Py 2.6
    def counts(self, only=Task):
        """
        Return a dictionary mapping each state name into the count of
        tasks in that state. In addition, the following keys are defined:

        * `ok`:  count of TERMINATED tasks with return code 0

        * `failed`: count of TERMINATED tasks with nonzero return code

        * `total`: total count of managed tasks, whatever their state

        If the optional argument `only` is not None, tasks whose
        whose class is not contained in `only` are ignored.

        : param class only: Restrict counting to tasks of these classes.
        """
        assert only in self._counts.totals
        return dictproxy(self._counts.totals[only])


    def stats(self, only=None):
        """
        Please use :meth:`counts` instead.

        .. warning::

          This is deprecated since GC3Pie version 2.5.
        """
        warn("Deprecated method `Engine.stats()` called"
             " -- please use `Engine.counts()` instead",
             DeprecationWarning, stacklevel=2)
        if only is None:
            # adapt to use `.counts()` default
            return self.counts()
        else:
            return self.counts(only)

    # implement a Core-like interface, so `Engine` objects can be used
    # as substitutes for `Core`.

    # pylint: disable=unused-argument
    def free(self, task, **extra_args):
        """
        Proxy for `Core.free`, which see.
        """
        self._core.free(task)

    def submit(self, task, resubmit=False, targets=None, **extra_args):
        """
        Submit `task` at the next invocation of `progress`.

        The `task` state is reset using the task's own method
        `.redo()`, and then the task added to the collection of
        managed tasks.  Note that the use of `redo()` implies that
        only tasks in a terminal state can be resubmitted!

        The `targets` argument is only present for interface
        compatiblity with `Core.submit`:meth: but is otherwise
        ignored.
        """
        if resubmit:
            # since we are going to change the task's state, we need
            # to expunge it from the queues ...
            if task in self._managed:
                self.remove(task)
            task.redo()
        # ... and then add it again with the (possibly) new state
        return self.add(task)

    def update_job_state(self, *tasks, **extra_args):
        """
        Return list of *current* states of the given tasks.  States
        will only be updated at the next invocation of `progress`; in
        particular, no state-change handlers are called as a result of
        calling this method.
        """
        pass

    def fetch_output(self, task, output_dir=None,
                     overwrite=False, changed_only=True, **extra_args):
        """
        Enqueue task for later output retrieval.

        .. warning:: FIXME

          The `output_dir`, `overwrite`, and `changed_only` parameters
          are currently ignored.
        """
        self.add(task)

    def kill(self, task, **extra_args):
        """
        Schedule a task for killing on the next `progress` run.
        """
        self._managed.requeue(task, 'kill')

    def peek(self, task, what='stdout', offset=0, size=None, **extra_args):
        """
        Proxy for `Core.peek` (which see).
        """
        return self._core.peek(task, what, offset, size, **extra_args)

    def close(self):
        """
        Call explicilty finalize methods on relevant objects
        e.g. LRMS
        """
        self._core.close()

    # Wrapper methods around `Core` to access the backends directly
    # from the `Engine`.

    # pylint: disable=missing-docstring

    @utils.same_docstring_as(Core.select_resource)
    def select_resource(self, match):
        return self._core.select_resource(match)

    @utils.same_docstring_as(Core.get_resources)
    def get_resources(self):
        return self._core.get_resources()

    @utils.same_docstring_as(Core.get_backend)
    def get_backend(self, name):
        return self._core.get_backend(name)


class BgEngine(object):
    """
    Run a GC3Pie `Engine`:class: instance in the background.

    A `BgEngine` exposes the same interface as a regular `Engine`
    class, but proxies all operations for asynchronous execution by
    the wrapped `Engine` instance.  In practice, this means that all
    invocations of `Engine` operations on a `BgEngine` always succeed:
    errors will only be visible in the background thread of execution.
    """
    def __init__(self, lib, *args, **kwargs):
        """
        Initialize an instance of class `BgEngine`:class:.

        :param str lib:
            framework to use for background thread scheduling;
            any value supported by `gc3libs.utils.get_scheduler_and_lock_factory`:func:
            (which see) is allowed here.

        :param args:
            Either a single `Engine`:class: instance, or a list of positional
            arguments to pass to the `Engine`:class: constructor.  In the former
            case, the instance should be the one and only argument (after `lib`).

        :param kwargs:
            Keyword arguments to forward to the `Engine`:class: constructor
            (unless a pre-built `Engine` instance is passed to `BgEngine`,
            in which case no keyword arguments are allowed.)
        """
        sched, lock = utils.get_scheduler_and_lock_factory(lib)
        self._scheduler = sched()
        self.running = False

        # a queue for Engine ops
        self._queue = []
        self._queue_locked = lock()

        # queues for before/after `Engine.progress()` triggers
        self._after_progress_triggers = []
        self._after_progress_triggers_locked = lock()
        self._before_progress_triggers = []
        self._before_progress_triggers_locked = lock()

        assert len(args) + len(kwargs) > 0, (
            "`BgEngine()` must be called"
            " either with an `Engine` instance as second (and last) argument,"
            " or with a set of parameters to pass on to the `Engine` constructor.")
        if len(args) == 1 and isinstance(args[0], gc3libs.core.Engine):
            # first (and only!) arg is an `Engine` instance, use that
            self._engine = args[0]
        else:
            # use supplied parameters to construct an `Engine`
            self._engine = gc3libs.core.Engine(*args, **kwargs)

        # no result caching until an update is really performed
        self._progress_last_run = 0


    #
    # control main loop scheduling
    #

    def start(self, interval):
        """
        Start triggering the main loop at the given `interval` frequency.

        :param gc3libs.quantity.Duration interval:
          Time span between successive calls of `_perform`:meth:
        """
        self.running = True
        self._scheduler.add_job(
            self._perform,
            'interval', seconds=(interval.amount(Duration.s)))
        self._scheduler.start()
        gc3libs.log.info(
            "Started background execution of Engine %s every %s",
            self._engine, interval)


    def stop(self, wait=False):
        """
        Stop background execution of the main loop.

        Call `start`:meth: to resume running.

        :param bool wait:
          When ``True``, wait until all pending actions
          on the background thread have been completed.
        """
        gc3libs.log.info(
            "Stopping background execution of Engine %s ...", self._engine)
        self.running = False
        self._scheduler.shutdown(wait)


    def _perform(self):
        """
        Main loop: runs in a background thread after `start`:meth: has
        been called.

        There are two tasks that this loop performs:

        - Execute any queued engine commands.

        - Run `Engine.progress()` to ensure that GC3Pie tasks are updated.
        """
        gc3libs.log.debug("%s: _perform() started", self)
        self.__run_delayed_operations()
        self.__run_before_triggers()
        self.__run_engine_progress()
        self.__run_after_triggers()

    def __run_delayed_operations(self):
        # quickly grab a local copy of the command queue, and
        # reset it to the empty list -- we do not want to hold
        # the lock on the queue for a long time, as that would
        # make the API unresponsive
        with self._queue_locked:
            queue = self._queue
            self._queue = []
        self.__run_hooks(queue)

    def __run_before_triggers(self):
        with self._before_progress_triggers_locked:
            before_progress_triggers = self._before_progress_triggers
            self._before_progress_triggers = []
        self.__run_hooks(before_progress_triggers)

    def __run_after_triggers(self):
        with self._after_progress_triggers_locked:
            after_progress_triggers = self._after_progress_triggers
            self._after_progress_triggers = []
        self.__run_hooks(after_progress_triggers)

    @staticmethod
    def __run_hooks(queue):
        """
        Call all the functions listed in `queue`, in the order given.
        Any exceptions raised will be logged at a WARNING level but
        otherwise ignored.
        """
        for func, args, kwargs in queue:
            gc3libs.log.debug(
                "Executing delayed call %s(*%r, **%r) ...",
                func.__name__, args, kwargs)
            try:
                func(*args, **kwargs)
            except Exception as err:  # pylint: disable=broad-except
                gc3libs.log.warning(
                    "Got '%s' while executing delayed call %s(*%r, **%r): %s",
                    err.__class__.__name__,
                    func.__name__, args, kwargs,
                    err, exc_info=__debug__)

    def __run_engine_progress(self):
        """
        Call the `.progress()` method of the wrapped `Engine` instance.
        """
        gc3libs.log.debug(
            "%s: calling `progress()` on Engine %s ...",
            self, self._engine)
        # pylint: disable=broad-except
        try:
            self._engine.progress()
            self._progress_last_run = time.time()
        except Exception as err:
            gc3libs.log.warning(
                "Ignoring '%s' error,"
                "  occurred while running"
                " `Engine.progress()` in the background: %s",
                err.__class__.__name__, err, exc_info=__debug__)
        gc3libs.log.debug("%s: _perform() done", self)


    def trigger_before_progress(self, func, *args, **kwargs):
        """
        Call a function *before* running `Engine.progress()` in the main loop.
        Exceptions raised during the call will be logged at WARNING level but
        otherwise ignored.

        The function call will be triggered only *once* at the next run of the
        main loop; it will not be fired repeatedly at every re-run of the main
        loop.

        Any suppplemental positional arguments or keyword-arguments that are
        supplied will be passed unchanged to the trigger function.
        """
        with self._before_progress_triggers_locked:
            self._before_progress_triggers.append((func, args, kwargs))


    def trigger_after_progress(self, func, *args, **kwargs):
        """
        Call a function *after* running `Engine.progress()` in the main loop.
        Exceptions raised during the call will be logged at WARNING level but
        otherwise ignored.

        The function call will be triggered only *once* at the next run of the
        main loop; it will not be fired repeatedly at every re-run of the main
        loop.

        Any suppplemental positional arguments or keyword-arguments that are
        supplied will be passed unchanged to the trigger function.
        """
        with self._after_progress_triggers_locked:
            self._after_progress_triggers.append((func, args, kwargs))


    @staticmethod
    def at_most_once_per_cycle(fn):  # pylint: disable=invalid-name
        """
        Ensure the decorated function is not executed more than once per
        each poll interval.

        Cached results are returned instead, if `Engine.progress()` has
        not been called in between two separate invocations of the wrapped
        function.

        .. warning::

          *Keyword arguments are ignored when doing a lookup* for
          previously-cached function results. This means that the
          following expressions might all return the same cached
          value::

            f(), f(foo=1), f(bar=2, baz='a')
        """
        # pylint: disable=missing-docstring,protected-access
        @functools.wraps(fn)
        def wrapper(self, *args, **kwargs):
            # no caching if the main loop is not running
            if not self._progress_last_run:
                return fn(self, *args, **kwargs)
            else:
                key = (fn, tuple(id(arg) for arg in args))
                try:
                    update = (self._cache_last_updated[key] < self._progress_last_run)
                except AttributeError:
                    self._cache_last_updated = defaultdict(float)
                    self._cache_value = {}
                    update = True
                if update:
                    self._cache_value[key] = fn(self, *args)
                    self._cache_last_updated[key] = time.time()
                # gc3libs.log.debug("%s(%s, ...): Using cached value '%s'",
                #                  fn.__name__, obj, obj._cache_value[key])
                return self._cache_value[key]
        return wrapper


    #
    # Engine interface
    #

    def add(self, task):
        """Proxy to `Engine.add`:meth: (which see)."""
        if self.running:
            with self._queue_locked:
                self._queue.append((self._engine.add, (task,), {}))
        else:
            self._engine.add(task)


    def close(self):
        """Proxy to `Engine.close`:meth: (which see)."""
        if self.running:
            with self._queue_locked:
                self._queue.append((self._engine.close, tuple(), {}))
        else:
            self._engine.close()


    def counts(self, only=Task):
        """Proxy to `Engine.counts`:meth: (which see)."""
        return self._engine.counts(only)


    def fetch_output(self, task, output_dir=None,
                     overwrite=False, changed_only=True, **extra_args):
        """Proxy to `Engine.fetch_output`:meth: (which see)."""
        if self.running:
            with self._queue_locked:
                self._queue.append((self._engine.fetch_output,
                                    (task, output_dir, overwrite, changed_only),
                                    extra_args))
        else:
            self._engine.fetch_output(task, output_dir, overwrite,
                                      changed_only, **extra_args)


    def find_task_by_id(self, task_id):
        """Proxy to `Engine.find_task_by_id`:meth: (which see)."""
        return self._engine.find_task_by_id(task_id)


    def free(self, task, **extra_args):
        """Proxy to `Engine.free`:meth: (which see)."""
        if self.running:
            with self._queue_locked:
                self._queue.append((self._engine.free, (task,), extra_args))
        else:
            self._engine.free(task, **extra_args)


    def get_resources(self):
        """Proxy to `Engine.get_resources`:meth: (which see)."""
        return self._engine.get_resources()


    def get_backend(self, name):
        """Proxy to `Engine.get_backend`:meth: (which see)."""
        return self._engine.get_backend(name)


    def iter_tasks(self):
        """
        Proxy to `Engine.iter_tasks`:meth: (which see).
        """
        return self._engine.iter_tasks()


    def kill(self, task, **extra_args):
        """Proxy to `Engine.kill`:meth: (which see)."""
        if self.running:
            with self._queue_locked:
                self._queue.append((self._engine.kill, (task,), extra_args))
        else:
            self._engine.kill(task, **extra_args)


    def peek(self, task, what='stdout', offset=0, size=None, **extra_args):
        """Proxy to `Engine.peek`:meth: (which see)."""
        if self.running:
            with self._queue_locked:
                self._queue.append((self._engine.peek,
                                    (task, what, offset, size), extra_args))
        else:
            self._engine.peek(task, what, offset, size, **extra_args)


    def progress(self):
        """
        Proxy to `Engine.progress`.

        If the background thread is already running, this is a no-op,
        as progressing tasks is already taken care of by the
        background thread.  Otherwise, just forward the call to the
        wrapped engine.
        """
        if self.running:
            pass
        else:
            self._engine.progress()


    def remove(self, task):
        """Proxy to `Engine.remove`:meth: (which see)."""
        if self.running:
            with self._queue_locked:
                self._queue.append((self._engine.remove, (task,), {}))
        else:
            self._engine.remove(task)


    def select_resource(self, match):
        """Proxy to `Engine.select_resource`:meth: (which see)."""
        if self.running:
            with self._queue_locked:
                self._queue.append((self._engine.select_resource, (match,), {}))
        else:
            self._engine.select_resource(match)


    def stats(self, only=None):
        """Proxy to `Engine.stats`:meth: (which see)."""
        return self._engine.stats(only)


    def submit(self, task, resubmit=False, targets=None, **extra_args):
        """Proxy to `Engine.submit`:meth: (which see)."""
        if self.running:
            with self._queue_locked:
                self._queue.append((self._engine.submit, (task, resubmit, targets), extra_args))
        else:
            self._engine.submit(task, resubmit, targets, **extra_args)


    def update_job_state(self, *tasks, **extra_args):
        """Proxy to `Engine.update_job_state`:meth: (which see)."""
        if self.running:
            with self._queue_locked:
                self._queue.append((self._engine.update_job_state, tasks, extra_args))
        else:
            self._engine.update_job_state(*tasks, **extra_args)



from __future__ import absolute_import, print_function, unicode_literals
from builtins import range
from past.builtins import basestring
import bz2
import json
import os
import re
import sys
import time
import tarfile
from multiprocessing import Manager, Pool, Process, current_process
from zipfile import ZipFile, is_zipfile

import pandas
from vaderSentiment.vaderSentiment import SentimentIntensityAnalyzer

analyzer = SentimentIntensityAnalyzer()


def parse_magic_ass_json(data):
    """
    The JSON format of twitter JSON files is JSONL-ish, parse it to lines of tweets (kv pairs)
    :param data: twitter json file
    :return: list of tweets (as kv pairs)
    """
    result = []
    print('parsing json (size: {0})'.format(len(data)))
    try:
        data = bz2.decompress(data)
        print('decompressed data to {0}'.format(len(data)))
    except IOError:
        pass
    for line in data.split('\n'):
        try:
            result.append(json.loads(line))
        except ValueError:
            pass
    return result


def extract_tweets(lines):
    """
    Extract the sentiment from a bunch of tweets
    :param lines: buch of lines containing kv pairs of twitter data
    :return: mean of negative, positive, neutral and compound sentiment in the lines
    """
    skipped = 0
    result = []
    for tweet in lines:
        if 'text' in tweet:
            scores = analyzer.polarity_scores(tweet.get('text'))
        elif 'body' in tweet:
            scores = analyzer.polarity_scores(tweet.get('body'))
        else:
            skipped += 1
            continue
        result.append([scores.get('neg'), scores.get('pos'), scores.get('neu'), scores.get('compound')])
    print('{0} tweets with no text/body field ({1} total), they have been skipped'.format(skipped, len(lines)))
    df = pandas.DataFrame(result, columns=['negative', 'positive', 'neutral', 'compound'])
    return df['negative'].mean(), df['positive'].mean(), df['neutral'].mean(), df['compound'].mean()


def process_twitter_json_file(archive_path, y, m, fq, wq):
    """
    Worker that processes a single twitter json file from the archive   
    :param archive_path: name of the archive
    :param y: cached year value
    :param m: cached month value
    :param fq: queue to read from
    :param wq: queue to write to
    """

    print('Starting process_twitter_json_file named {0}'.format(current_process().name))

    while True:
        archive_file = fq.get()
        if isinstance(archive_file, basestring) and archive_file == 'EOP':
            break
        else:
            print('going to extract archive file {0} from {1}'.format(archive_file.name, archive_path))
            with tarfile.open(archive_path) as tf:
                extracted_archive_file = tf.extractfile(archive_file)
                if not extracted_archive_file.closed:
                    json_data = extracted_archive_file.read()
                    data = parse_magic_ass_json(json_data)
                    extracted_archive_file.close()
                    neg, pos, neu, com = extract_tweets(data)
                    if len(archive_file.name.split('/')) == 5:
                        yr, mt, day, hour, p = archive_file.name.split('/')
                    elif len(archive_file.name.split('/')) == 4 and year:
                        yr = y
                        mt, day, hour, p = archive_file.name.split('/')
                    elif len(archive_file.name.split('/')) == 3 and year and month:
                        yr = y
                        mt = m
                        day, hour, p = archive_file.name.split('/')
                    else:
                        raise ValueError('cannot infer date from filename and/or directory structure')
                    wq.put('{y}.{m}.{d}.{h}.{p};{neg};{pos};{neu};{com};\n'.format(y=yr,
                                                                                   m=mt,
                                                                                   d=day,
                                                                                   p=p.split('.')[0],
                                                                                   h=hour,
                                                                                   neg=neg,
                                                                                   pos=pos,
                                                                                   neu=neu,
                                                                                   com=com))
                else:
                    print('file pointer not opened, something went wrong for {0} in {1}, skipping..,'.format(archive_file.name, archive_path))


def mp_writer(o, wq):
    """
    Multiprocess listener. Listens to message on Queue and writes them to file
    :param o: name of the output file
    :param wq: queue to read from 
    """
    with open(o, 'wb') as fp:
        while True:
            m = wq.get()
            if m == 'EOP':
                fp.flush()
                break
            fp.write(m)
            fp.flush()


def queue_archive_files(archive_path, fq):
    """
    Extract all data and process everything :)
    :param archive_path: name of the archive
    :param fq: puts the file names on the queue 
    """
    if is_zipfile(archive_path):
        with ZipFile(archive_path, 'r') as archive:
            for archive_file in archive.namelist():
                if not archive_file.endswith('/') and ZipFile.getinfo(archive, archive_file).file_size > 0:
                    fq.put(archive_file)
    elif tarfile.is_tarfile(archive_path):
        with tarfile.open(archive_path) as archive:
            for archive_file in archive:
                if archive_file.isreg() and archive_file.size > 0:
                    fq.put(archive_file)


if __name__ == "__main__":
    # wait {timeout} seconds for our file to become available
    timeout = 360
    if len(sys.argv) == 3:
        while not os.path.exists(sys.argv[1]):
            timeout -= 1
            if timeout > 0:
                print('file {0} not available yet, waiting {1} more seconds'.format(sys.argv[1], timeout))
            else:
                print('timout exceeded, terminating program, please check configuration for issues')
                exit(-2)
            time.sleep(1)
        year = None
        month = None
        match = re.search(r'([0-9]{4})-([0-9]{2})', sys.argv[1])
        if len(match.groups()) == 2:
            year = match.group(1)
            month = match.group(2)
            print('found something resembling a year/month thingy ({0}-{1})in the file name, caching it.'.format(year, month))
        print('setting up multiprocessing')
        manager = Manager()
        file_queue = manager.Queue()
        writer_queue = manager.Queue()
        print('Setup our file indexer')
        indexer = Process(target=queue_archive_files, args=(sys.argv[1], file_queue,))
        indexer.start()
        print('Setup our output writer')
        writer = Process(target=mp_writer, args=('{fn}.csv'.format(fn=os.path.basename(sys.argv[1])), writer_queue,))
        writer.start()
        print('Setup file processors and start crunching')
	jobs = []
	for index in range(0,int(sys.argv[2])):
	    jobs.append(Process(target=process_twitter_json_file, args=(sys.argv[1], year, month, file_queue, writer_queue,)))
   	for job in jobs:
	    job.start()
	    print('Started process {0} in pid {1}'.format(job.name,job.pid))
	print('wait for our indexer to finish')
        indexer.join()
        print('add terminators to index file queue (as last element)')
        for job in jobs:
            file_queue.put('EOP')
        print('wait for our processors to finnish')
        for job in jobs:
            job.join()
        print('and now add terminator for writer and wait to finish')
        writer_queue.put('EOP')
        writer.join()
        print('Elvis has left the building')
    else:
        print('Invalid argument count, should only be one single archive and a cpu core count.')
        exit(-1)





#! /usr/bin/env python
#
"""
Test persistence backends.
"""
# Copyright (C) 2011, 2012, 2019,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
from builtins import range
from builtins import object
__docformat__ = 'reStructuredText'

# stdlib imports
import os
import shutil
from tempfile import NamedTemporaryFile, mkdtemp

# 3rd party imports
import pytest

sqlalchemy = pytest.importorskip("sqlalchemy")
sql = pytest.importorskip("sqlalchemy.sql")
sqlfunc = sqlalchemy.func

# GC3Pie imports
import gc3libs
from gc3libs import Run, Task
import gc3libs.workflow

import gc3libs.exceptions
from gc3libs.persistence import make_store, Persistable
from gc3libs.persistence.accessors import GET
from gc3libs.persistence.serialization import DEFAULT_PROTOCOL
from gc3libs.persistence.idfactory import IdFactory
from gc3libs.persistence.filesystem import FilesystemStore
from gc3libs.persistence.sql import SqlStore
from gc3libs.url import Url, UrlKeyDict


@pytest.mark.parametrize("cls", (SqlStore, FilesystemStore))
def test_store_ctor_with_extra_arguments(cls, tmpdir):
    """Test if `Store`:class: classess accept extra keyword arguments
    without complaining.

    This test will ensure that any `Store` class can be called with
    arguments which are valid only on other `Store` classes.
    """
    args = {
        'url': 'sqlite:///%s/test.sqlite' %
        os.path.abspath(str(tmpdir)),
        'table_name': 'store',
        'create': True,
        'directory': str(tmpdir),
        'idfactory': IdFactory(),
        'protocol': DEFAULT_PROTOCOL,
        'extra_fields': {
            sqlalchemy.Column(
                'extra',
                sqlalchemy.TEXT()): lambda x: "test"},
    }

    cls(**args)


def test_eq_persisted_objects():
    """Test that the comparison of two persisted objects are the same
    whenever `persistent_id` attribute is the same"""
    a = Persistable()
    b = Persistable()
    c = Persistable()
    a.persistent_id = '1'
    b.persistent_id = '1'
    c.persistnet_id = '2'
    assert a == b
    assert a is not b
    assert a != c
    L = [a, b, c]
    L.remove(a)
    L.remove(a)
    assert len(L) == 1

def test_eq_non_persisted_objects():
    """Test that the comparison of two not (yet) persisted objects falls
    back to default Python comparison"""
    a = Persistable()
    b = Persistable()
    assert a != b
    assert a is not b


# for testing basic functionality we do no need fully-fledged GC3Pie
# objects; let's define some simple make-do's.

class SimplePersistableObject(Persistable):

    def __init__(self, x):
        self.value = x

    def __eq__(self, other):
        return (self.value == other.value)


class SimplePersistableList(list, Persistable):

    """
    Add a `__dict__` to `list`, so that creating a `persistent_id`
    entry on an instance works.
    """
    pass


class SimpleTask(Task):
    # XXX: do we need this? why not instanciating `Task` directly?
    pass


class NonPersistableClassWithSlots(object):

    """
    A class to test that persistence of classes with `__slots__`.

    This and the `PersistableClassWithSlots`:class: are classes with
    `__slots__` attribute to check that persistence works also with
    this kind of classes. Usually you just need to use a binary
    protocol for pickle.

    This class will raise an error because `__slots__` does not
    contain `persistent_id`. The `PersistableClassWithSlots`:class:
    class instead, will work as expected.

    Check for instance:
    http://stackoverflow.com/questions/3522765/python-pickling-slots-error
    """

    __slots__ = ["attr", ]

    def __init__(self, attr):
        self.attr = attr


class PersistableClassWithSlots(NonPersistableClassWithSlots):
    __slots__ = ["attr", "persistent_id"]


class MyChunkedParameterSweep(gc3libs.workflow.ChunkedParameterSweep):

    def new_task(self, param, **extra_args):
        return Task(**extra_args)


class MyStagedTaskCollection(gc3libs.workflow.StagedTaskCollection):

    def stage0(self):
        return Task()


# def _run_generic_tests(store):
#     _generic_persistence_test(store)
#     _generic_nested_persistence_test(store)
#     _generic_persist_classes_with_slots(store)
#     _save_different_objects_separated(store)


class GenericStoreChecks(object):

    """
    A suite of tests for the generic `gc3libs.persistence.Store` interface.
    """

    def test_generic_persistence(self):
        """
        Test basic functionality of the persistence subsystem.
        """
        # check that basic save/load works
        obj = SimplePersistableObject('GC3')
        id = self.store.save(obj)
        del obj
        obj = self.store.load(id)
        assert obj.value == 'GC3'

    def test_duplicated_save(self):
        """
        Check that if an object is already on the db a call to
        `self.store.save` will not save a duplicate of the object, but
        it will override the old one.
        """
        obj = SimplePersistableObject('GC3')

        id1 = self.store.save(obj)
        id2 = self.store.save(obj)
        assert id1 == id2

    def test_ids_are_not_duplicated(self):
        """
        check that IDs are never duplicate
        """
        ids = []
        for i in range(10):
            ids.append(self.store.save(SimplePersistableObject(str(i))))
        assert len(ids) == len(set(ids))

    def test_list_method(self):
        """Test the `list` method of the `SqlStore` class"""
        num_objs = 10
        for i in range(num_objs):
            self.store.save(SimplePersistableObject('Object %d' % i))

        assert len(self.store.list()) == num_objs

    def test_remove_method(self):
        """
        Test remove method of a generic `Store` class
        """
        # Removing objects
        obj = SimplePersistableObject('GC3')
        id = self.store.save(obj)
        self.store.remove(id)

        with pytest.raises(gc3libs.exceptions.LoadError):
            obj = self.store.load(id)

    def test_replace_method(self):
        """Test the `replace` method of the `SqlStore` class"""
        # 1) save a new object
        obj = SimplePersistableObject("Original")
        id_ = self.store.save(obj)

        # 2) change it and replace it
        obj.x = "Updated"
        self.store.replace(obj.persistent_id, obj)
        assert id_ == obj.persistent_id

        # 3_ load it again and check if it has been updated
        obj2 = self.store.load(id_)
        assert obj2.x == "Updated"

    @pytest.mark.skip(reason="FIXME: Test code needs to be checked!")
    def test_persist_classes_with_slots(self):

        # FIXME: check this code!
        obj = PersistableClassWithSlots('GC3')
        assert obj.attr == 'GC3'
        id_ = self.store.save(obj)
        del obj
        obj2 = self.store.load(id_)
        assert obj2.attr == 'GC3'

        obj2 = NonPersistableClassWithSlots('GC3')
        try:
            self.store.save(obj2)
            raise AssertionError("We shouldn't reach this point")
        except AttributeError:
            pass

    def test_disaggregate_persistable_objects(self):
        """
        Check that `Persistable` instances are saved separately from
        their containers.

        Since we want to save `Task`s separately from the containing
        object (e.g. `SessionBasedScript`), this test will try to save a
        generic persisted container which contain another persisted and
        check that they are saved on different items.

        This is done by checking that all the objects have a unique
        `persistent_id` attribute.
        """
        container = SimplePersistableList()
        container.append(SimplePersistableObject('MyJob'))
        container_id = self.store.save(container)

        # check that IDs have been assigned
        assert hasattr(container, 'persistent_id')
        assert hasattr(container[0], 'persistent_id')

        # check that IDs are distinct
        assert container_id == container.persistent_id
        objid = container[0].persistent_id
        assert objid != container_id

        # check that loading the container re-creates the same contained object
        del container
        container = self.store.load(container_id)
        obj = self.store.load(objid)
        assert obj == container[0]

        # return objects for further testing
        return (container_id, objid)

    def test_persist_urlkeydict(self):
        """
        Test that we can persist GC3Pie's `UrlKeyDict` classes.
        """
        a = Task(attr=UrlKeyDict({'/tmp/1': 1, '/tmp/2': 2}))
        id_ = self.store.save(a)
        b = self.store.load(id_)
        assert b.attr == a.attr


    def test_persist_urlvaluedict(self):
        """
        Test that we can persist GC3Pie's `UrlValueDict` classes.
        """
        a = Task(attr=UrlKeyDict({'foo': '/tmp/1', 'bar': '/tmp/2'}))
        id_ = self.store.save(a)
        b = self.store.load(id_)
        assert b.attr == a.attr


@pytest.mark.parametrize("task", (Task(),
        gc3libs.workflow.TaskCollection(tasks=[Task(), Task()]),
        gc3libs.workflow.SequentialTaskCollection([Task(), Task()]),
        MyStagedTaskCollection(),
        gc3libs.workflow.ParallelTaskCollection(tasks=[Task(), Task()]),
        MyChunkedParameterSweep(1, 20, 1, 5),
        gc3libs.workflow.RetryableTask(Task())))
def test_task_objects_buggy(task):
    """
    Test that all `Task`-like objects are persistable
    """
    with NamedTemporaryFile(prefix='gc3libs.', suffix='.tmp') as tmp:
        store = make_store("sqlite://%s" % tmp.name)
        id = store.save(task)
        store.load(id)


class TestFilesystemStore(GenericStoreChecks):

    @pytest.fixture(autouse=True)
    def setUp(self):
        from gc3libs.persistence.filesystem import FilesystemStore
        self.tmpdir = mkdtemp(prefix='gc3libs.', suffix='.tmp.d')
        self.store = FilesystemStore(self.tmpdir)

        yield

        shutil.rmtree(self.tmpdir)

    # XXX: there's nothing which is `FilesystemStore`-specific here!
    def check_file_accesssystemstorage_pickler_class(self):
        """
        Check that `Persistable` instances are saved saparately.

        Namely, if you want to save two independent objects but one of
        them has a reference to the other, the standard behavior of
        Pickle is to save a copy of the contained object into the same
        file of the containing object.

        The FilesystemStorage.Pickler class is aimed to avoid this.
        """
        obj1 = SimplePersistableObject('parent')
        obj2 = SimplePersistableObject('children')
        obj1.children = obj2
        id2 = self.store.save(obj2)
        id1 = self.store.save(obj1)
        del obj1
        del obj2
        obj1 = self.store.load(id1)
        obj2 = self.store.load(id2)
        assert obj1.children.value == 'children'
        assert obj2.value == 'children'
        assert obj1.children == obj2

    def test_disaggregate_persistable_objects(self):
        """
        Check that `Persistable` instances are saved separately from
        their containers.

        In addition to the checks done in `GenericStoreChecks`, we
        also test that files have been created.
        """
        # std checks
        container_id, obj_id = super(
            TestFilesystemStore, self).test_disaggregate_persistable_objects()
        # check that files exist
        container_file = os.path.join(self.store._directory, str(container_id))
        assert os.path.exists(container_file)
        obj_file = os.path.join(self.store._directory, str(obj_id))
        assert os.path.exists(obj_file)


class SqlStoreChecks(GenericStoreChecks):

    """
    Extend `GenericStoreChecks` with additional SQL-related tests.

    In order for these tests to work, the class constructor must
    define these instance attributes:

    * `self.db_url`: SqlAlchemy URL to open the database
    * `self.store`: a `gc3libs.persistence.Store` instance to use for tests
    * `self.c`: a valid cursor for executing raw SQL queries.
    """

    def test_default_fields(self):
        app = gc3libs.Application(
            arguments=['/bin/true'],
            inputs=[],
            outputs=[],
            output_dir='/tmp',
            jobname='test_job_persistence')

        app.execution.state = Run.State.NEW
        app.execution.lrms_jobid = 1
        id_ = self.store.save(app)

        q = sql.select([
            self.store._tables.c.state
        ]).where(self.store._tables.c.id == id_)
        result = self.conn.execute(q)
        row = result.fetchone()

        assert row[0] == app.execution.state

    # the `jobname` attribute is optional in the `Application` ctor
    def test_persist_Application_with_no_job_name(self):
        app = gc3libs.Application(
            arguments=['/bin/true'],
            inputs=[],
            outputs=[],
            output_dir='/tmp')

        app.execution.state = Run.State.NEW
        app.execution.lrms_jobid = 1
        id_ = self.store.save(app)

        q = sql.select([
            self.store._tables.c.state
        ]).where(self.store._tables.c.id == id_)
        result = self.conn.execute(q)
        row = result.fetchone()
        assert row[0] == app.execution.state

    def test_sql_injection(self):
        """Test if the `SqlStore` class is vulnerable to SQL injection."""

        storetemp = make_store(
            self.db_url,
            table_name='sql_injection_test',
            extra_fields={
                sqlalchemy.Column(
                    'extra',
                    sqlalchemy.VARCHAR(
                        length=128)): GET.extra,
            })

        obj = SimpleTask()
        # obligatory XKCD citation ;-)
        # Ric, you can't just "cite" XKCD without inserting a
        # reference: http://xkcd.com/327/
        obj.extra = "Robert'); DROPT TABLE sql_injection_test; --"
        id_ = storetemp.save(obj)
        obj2 = storetemp.load(id_)
        self.conn.execute('drop table sql_injection_test')
        assert obj.execution.state == obj2.execution.state

    def test_sql_save_load_extra_fields(self):
        """
        Test if `SqlStore` reads and writes extra columns.
        """
        # extend the db
        self.conn.execute('alter table `%s` add column extra varchar(256)'
                          % self.store.table_name)

        # re-build store, as the table list is read upon `__init__`
        self.store = self._make_store(
            extra_fields={
                sqlalchemy.Column(
                    'extra',
                    sqlalchemy.VARCHAR(
                        length=128)): GET.foo.value,
            })

        # if this query does not error out, the column is defined
        q = sql.select([sqlfunc.count(self.store._tables.c.extra)]).distinct()
        results = self.conn.execute(q)
        rows = results.fetchall()
        assert len(rows) == 1

        # create and save an object
        obj = SimplePersistableObject('an object')
        obj.foo = SimplePersistableObject('an attribute')
        id_ = self.store.save(obj)

        # check that the value has been saved
        q = sql.select([self.store._tables.c.extra]).where(
            self.store._tables.c.id == id_)

        # Oops, apparently the store.save call will close our
        # connection too.
        self.conn = self.store._engine.connect()
        results = self.conn.execute(q)
        rows = results.fetchall()
        assert len(rows) == 1
        assert rows[0][0] == obj.foo.value

    @pytest.mark.skip(reason="FIXME: Check if test is still valid")
    def test_sql_error_if_no_extra_fields(self):
        """
        Test if `SqlStore` reads and writes extra columns.
        """
        # re-build store with a non-existent extra column
        with pytest.raises(AssertionError):
            self._make_store(
                extra_fields={
                    sqlalchemy.Column(
                        'extra',
                        sqlalchemy.VARCHAR(
                            length=128)): (
                                lambda arg: arg.foo.value)})


class ExtraSqlChecks(object):

    """
    `SqlStore` that depend on the setup of a non-default table.
    """

    def _make_store(self, url, **kwargs):
        self.db_url = url
        self.store = make_store(url, **kwargs)
        self.c = self.store._engine.connect()

    def test_sql_create_extra_fields(self):
        """
        Test if `SqlStore` creates extra columns.
        """

        # extend the db
        self._make_store(
            self.db_url,
            extra_fields={
                sqlalchemy.Column(
                    'extra',
                    sqlalchemy.VARCHAR(
                        length=128)): (
                    lambda arg: arg.foo.value)})

        # if this query does not error out, the column is defined
        q = sql.select([sqlfunc.count(self.store._tables.c.extra)]).distinct()
        results = self.conn.execute(q)
        rows = results.fetchall()
        assert len(rows) == 1

        # create and save an object
        obj = SimplePersistableObject('an object')
        obj.foo = SimplePersistableObject('an attribute')
        id_ = self.store.save(obj)

        # check that the value has been saved
        q = sql.select([self.store._tables.c.extra]).where(
            self.store._tables.c.id == id_)
        # self.c.execute("select extra from %s where id=%d"
        #                % (self.store.table_name, id_))
        results = self.conn.execute(q)
        rows = results.fetchall()
        assert len(rows) == 1
        assert rows[0][0] == obj.foo.value


class TestSqliteStore(SqlStoreChecks):

    """Test SQLite backend."""

    @classmethod
    def setup_class(cls):
        # skip SQLite tests if no SQLite module present (Py 2.4)
        sqlite3 = pytest.importorskip("sqlite3")

    @pytest.fixture(autouse=True)
    def setUp(self):
        with NamedTemporaryFile(prefix='gc3libs.', suffix='.tmp') as tmp:
            self.tmpfile = tmp.name
            self.db_url = Url('sqlite://%s' % self.tmpfile)
            self.store = self._make_store()

            # create a connection to the database
            self.conn = self.store._engine.connect()

            yield
            self.conn.close()

    def _make_store(self, **kwargs):
        return make_store(self.db_url, **kwargs)


class TestSqliteStoreWithAlternateTable(TestSqliteStore):

    """Test SQLite backend with a different table name."""

    @classmethod
    def setup_class(cls):
        # skip SQLite tests if no SQLite module present (Py 2.4)
        sqlite3 = pytest.importorskip("sqlite3")

    def _make_store(self, **kwargs):
        return make_store(self.db_url, table_name='another_store', **kwargs)

    def test_alternate_table_not_empty(self):
        obj = SimplePersistableObject('an object')
        self.store.save(obj)

        results = self.conn.execute(sql.select([self.store._tables.c.id]))
        rows = results.fetchall()
        assert len(rows) > 0


class TestMysqlStore(SqlStoreChecks):

    @classmethod
    def setup_class(cls):
        # we skip MySQL tests if no MySQLdb module is present
        MySQLdb = pytest.importorskip("MySQLdb")

    @classmethod
    def teardown_class(cls):
        pass

    @pytest.fixture(autouse=True)
    def setUp(self):
        # generate random table name
        from string import ascii_letters as letters
        import random
        self.table_name = 'test_' + (''.join(random.choice(letters) for _ in range(10)))

        try:
            self.db_url = Url('mysql://gc3user:gc3pwd@localhost/gc3')
            self.store = make_store(self.db_url, table_name=self.table_name)
        except sqlalchemy.exc.OperationalError:
            pytest.mark.skip("Cannot connect to MySQL database.")

        # create a connection to the database
        self.conn = self.store._engine.connect()

        yield

        self.conn.execute('drop table `%s`' % self.table_name)
        self.conn.close()
        # self.c.close()
        # os.remove(self.tmpfile)

    def _make_store(self, **kwargs):
        return make_store(self.db_url, table_name=self.table_name, **kwargs)


if __name__ == "__main__":
    pytest.main(["-v", __file__])



#! /usr/bin/env python

from __future__ import absolute_import, print_function
import sys
import fnmatch
import os
from optparse import OptionParser
import shutil

"""
1. new simulation: just generated
2. was already running: mixed with some results
3. pooluted by ggeotop

take *.asc as valid inputs also for job submission
folder pattern:
ggeotop.inpts
in/
out/

retrieve only out/
in/ remain unchaged

"""

dirs = ["in", "maps" "rec", "rad"]
# files = ["svf.asc","slp.asc", "asp.asc", "dem.asc"]

def search_and_list(input_folder):
    # agreed vlid input folder format:
    # ggeotop.inpts
    # in/
    # out/
    #
    for r,d,f in os.walk(input_folder):
        if "geotop.inpts" in f and "in" in d and "out" in d:
            print(r)

def main():
    """
    walk through the folder and remove
    N0*.asc
    out/
    *~
    """
    parser = OptionParser(usage="%prog [clean|list] [INPUT FOLDERs]")
    parser.add_option("-s", "--simulate", action="store_true", dest="simulate",
                      metavar="LEVEL",
                      help="Execute scrit in simulation mode"
                      " (default is FALSE).")

    parser.add_option("-f", "--full", action="store_true", dest="rec_also",
                      metavar="LEVEL",
                      help="Force removing also 'rec' folder")

    (options, args) = parser.parse_args(sys.argv[1:])

    if len(args) < 2:
        print(parser.usage)
        return 1

    command = args[0]
    input_folders = args[1:]
    
    if command == "list":
        for folder in input_folders:
            search_and_list(folder)
    elif command == "clean":
        for folder in input_folders:
            clean_folder(folder, simulate=options.simulate, rec_also=options.rec_also)
    else:
        print(parser.usage)

def clean_folder(input_folder, simulate, rec_also):
    
    filename_patterns = ["*~", "*.tgz"]
    dirname_patterns = ["tmp","*~"]

    if rec_also:
        dirname_patterns.append("rec")

    for r,d,f in os.walk(input_folder):
        
        skip = False
        
        for directory in dirs:
            if r.endswith(directory):
                skip = True

        if skip:
            continue

        # previous approach. keep for the record
        # folders_to_remove = [ folder for folder in d if folder in dirname_patterns ]
        # files_to_remove = [ fn for fn in f if fn in filename_patterns ]

        # start remove folders
        for folder in d:
            for p in dirname_patterns:
                if fnmatch.fnmatch(folder,p):
                    try:
                        # XXX: is this too strong assumption ?
                        # remove a folder 'out'
                        if simulate:
                            print("Removing folder %s" % os.path.join(r,folder))
                        else:
                            shutil.rmtree(os.path.join(r,folder), ignore_errors=True)
                            if folder == "rec":
                                # create an empty new one
                                os.mkdir(os.path.join(r,folder))
                    except Exception as x:
                        print("Failed deleting folder %s. Error %s. Message %s" % (folder,x.__class__,x.message))
                    break

        for fn in f:
            # try remove file matching pattern
            for p in filename_patterns:
                if fnmatch.fnmatch(fn,p):
                    try:
                        if simulate:
                            print("Removing file %s" % os.path.join(r,fn))
                        else:
                            os.remove(os.path.join(r,fn))
                    except Exception as x:
                        print("Failed deleting file %s. Error %s. Message %s" % (fn,x.__class__,x.message))
                    break


if __name__ == "__main__":
    sys.exit(main())
